# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Manipulate tag objects for Firmant.
'''


import os

from firmant import decorators
from firmant import parsers


class Tag(parsers.RstParsedObject):
    '''A tag is a means of categorizing objects.

    Attributes of :class:`Tag`:

        slug
           A unique string that identifies the tag.

        title
           A longer title used for identifying the tag to the user.

        subtitle
           An even longer description of the tag that may be displayed to the
           user in combination with `title`.

        content
           The content of the restructured text document.  This does not include
           the title information.

        posts
           A list of cross-referenced posts.  This will be blank until
           cross-referencing happens at a later point in time.

    .. note::

       All string attributes except `slug` may be ``''``.  Posts will be ``[]``
       until the cross-referencing chunk happens.

    .. doctest::

       >>> Tag(slug='foo')
       Tag(foo)

    '''

    # pylint: disable-msg=R0903

    __slots__ = ['slug', 'posts']

    _pubparts = [('content', 'fragment')
                   ,('title', 'title')
                   ,('subtitle', 'subtitle')
                   ]

    def __repr__(self):
        return 'Tag(%s)' % getattr(self, 'slug', None)

    @property
    def _attributes(self):
        '''Attributes that identify a tag object:

            slug
               The `slug` attribute of the :class:`Tag` object.

        .. doctest::

           >>> pprint(Tag(slug='firmantstuff')._attributes)
           {'slug': 'firmantstuff'}

        '''
        # We remove the 'rst' extension
        # pylint: disable-msg=E1101
        return {'slug': self.slug}


class TagParser(parsers.RstParser):
    r'''Parse all tags matching ``[-a-zA-Z0-9_.]+``.

    .. doctest::
       :hide:

       >>> environment['log'] = get_logger()

    .. doctest::

       >>> tp = TagParser(environment, objects)
       >>> environment, objects, (parser,) = tp(environment, objects)
       >>> pprint(parser(environment, objects)) #doctest: +ELLIPSIS
       ({...},
        {'tag': [Tag(bar), Tag(baz), Tag(foo), Tag(quux)]},
        [])

    '''

    type = 'tag'
    paths = '^[-a-zA-Z0-9_.]+\.rst$'
    cls = Tag

    @decorators.in_environment('settings')
    def root(self, environment):
        '''The directory under which all tag objects reside.
        '''
        settings = environment['settings']
        return os.path.join(settings.CONTENT_ROOT, settings.TAGS_SUBDIR)

    def rstparse(self, environment, objects, path, pieces):
        '''Use the parsed rst document to construct the necessary objects.
        '''
        attrs = {}
        attrs['slug'] = unicode(path[:-4])
        attrs['_pub'] = pieces['pub']
        objects[self.type].append(self.cls(**attrs))


def _setup(test):
    '''Setup the tests.
    '''
    from pysettings.settings import Settings
    from firmant import routing
    test.globs['settings'] = Settings({'CONTENT_ROOT': 'testdata/pristine'
                                      ,'TAGS_SUBDIR': 'tags'
                                      })
    urlmapper = routing.URLMapper('/path/to/output/dir', 'http://testurl/', [])
    test.globs['environment'] = {'settings': test.globs['settings']
                                ,'urlmapper': urlmapper
                                }
    test.globs['objects'] = {}



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''The Firmant application.

Orchestrates the high process of parsing, transforming, and writing content.
'''


import logging

from firmant import utils
from firmant.chunks import AbstractChunk
from firmant.routing import URLMapper
from firmant.utils import class_name


class Firmant(object):
    '''Perform a complete run from parsing through writing.

        >>> from pysettings.loaders import mod_to_settings
        >>> s = mod_to_settings('firmant.settings')
        >>> s.OUTPUT_DIR = outdir
        >>> f = Firmant(s)
        >>> f.log = Mock('log')
        >>> f()
        >>> pprint(f.objs) #doctest: +ELLIPSIS
        {'feed': [Feed(bar), Feed(baz), Feed(foo), Feed(quux)],
         'post': [Post(2009-12-31-party),
                  Post(2010-01-01-newyear),
                  Post(2010-02-01-newmonth),
                  Post(2010-02-02-newday),
                  Post(2010-02-02-newday2)],
         'static': [static_obj<testdata/pristine/static/images/88x31.png>],
         'staticrst': [staticrst_obj<about>,
                       staticrst_obj<empty>,
                       staticrst_obj<links>],
         'tag': [Tag(bar), Tag(baz), Tag(foo), Tag(quux)]}

    '''

    # pylint: disable-msg=R0903

    def __init__(self, settings):
        self.settings = settings
        self.urlmapper = URLMapper(settings.OUTPUT_DIR, settings.PERMALINK_ROOT,
                getattr(settings, 'URLS', None))
        self.log = logging.getLogger(class_name(self.__class__))

        self.objs = dict()
        self.writers = dict()
        self.chunks = list()
        self.env    = dict()
        self.env['log'] = self.log
        self.env['urlmapper'] = self.urlmapper
        self.env['settings'] = self.settings

        self.chunks.append(CheckURLConflicts())
        self.chunks.append(CreatePermalinks())
        self.chunks.append(CrossReference())

        for glob in settings.GLOBALS:
            glob = utils.get_obj(glob)
            self.chunks.append(glob(self.env, self.objs))

        # Setup parsers
        for parser in settings.PARSERS:
            parser = utils.get_obj(parser)
            if issubclass(parser, AbstractChunk):
                self.chunks.append(parser(self.env, self.objs))
            else:
                self.log.error(_("'%s' is not a parser.") % str(parser))

        # Setup writers
        for writer in self.settings.WRITERS:
            writer = utils.get_obj(writer)
            self.chunks.append(writer(self.env, self.objs))

    def __call__(self):
        while len(self.chunks):
            self.chunks.sort(key = lambda x: x.scheduling_order)
            chunk = self.chunks[0]
            self.env, self.objs, newchunks = chunk(self.env, self.objs)
            self.chunks = self.chunks[1:]
            for newchunk in newchunks:
                if newchunk.scheduling_order <= chunk.scheduling_order:
                    error = _('New chunk violates scheduling_order constraint')
                    self.log.error(error)
                else:
                    self.chunks.append(newchunk)


class CrossReference(AbstractChunk):
    '''A chunk that cross-references posts/tags/feeds.
    '''

    # TODO improve this to do arbitrary cross-references.
    # pylint: disable-msg=R0903

    def __call__(self, environment, objects):
        '''Cross reference tags and feeds
        '''
        # pylint: disable-msg=R0912
        for feed in objects.get('feed', []):
            feed.posts = list()
        for tag in objects.get('tag', []):
            tag.posts = list()
        for post in objects.get('post', []):
            to_delete = list()
            for i, ptag in enumerate(post.tags):
                seen = False
                for tag in objects.get('tag', []):
                    if tag.slug == ptag:
                        seen = True
                        tag.posts.append(post)
                        post.tags[i] = tag
                if not seen:
                    warning  = _("Tag '%s' referenced but not defined.")
                    warning %= ptag
                    environment['log'].warning(warning)
                    to_delete.append(i)
            for i in reversed(to_delete):
                del post.tags[i]
            to_delete = list()
            for i, pfeed in enumerate(post.feeds):
                seen = False
                for feed in objects.get('feed', []):
                    if feed.slug == pfeed:
                        seen = True
                        feed.posts.append(post)
                        post.feeds[i] = feed
                if not seen:
                    error  = _("Feed '%s' referenced but not defined.")
                    error %= pfeed
                    environment['log'].error(error)
                    to_delete.append(i)
            for i in reversed(to_delete):
                del post.feeds[i]
        return (environment, objects, [])

    scheduling_order = 700


class CheckURLConflicts(AbstractChunk):
    '''A chunk that warns of writers with conflicting URLs.
    '''

    # pylint: disable-msg=R0903

    def __call__(self, environment, objects):
        '''Ensure that the sets of URLs are disjoint.
        '''
        urls = dict()
        for key, inst in environment['urls'].items():
            for url in inst:
                if url is None:
                    warning  = _("Writer %s's URLs incompletely defined.")
                    warning %= key
                    environment['log'].warning(warning)
                elif url in urls:
                    warning  = _('Writers %s and %s conflict over %s.')
                    warning %= urls[url], key, url
                    environment['log'].warning(warning)
                else:
                    urls[url] = key
        return (environment, objects, [])

    scheduling_order = 600


class CreatePermalinks(AbstractChunk):
    '''For objects that declare attributes, add a permalink.

    Pre-existing permalinks will not be overwritten.

    '''

    # pylint: disable-msg=R0903

    def __call__(self, environment, objects):
        if 'settings' not in environment:
            environment['log'].error(_('Expected `settings` in environment.'))
            return
        if 'urlmapper' not in environment:
            environment['log'].error(_('Expected `urlmapper` in environment.'))
            return
        settings = environment['settings']
        urlmapper = environment['urlmapper']
        for typ, objlist in objects.items():
            for obj in objlist:
                if hasattr(obj, '_attributes') \
                        and (not hasattr(obj, 'permalink')
                             or obj.permalink is None):
                    # pylint: disable-msg=W0212
                    attrs = obj._attributes
                    attrs['type'] = typ
                    extension = settings.PERMALINK_EXTENSIONS[typ]
                    obj.permalink = urlmapper.url(extension, **attrs)
        return (environment, objects, [])

    scheduling_order = 300


def _setup(self):
    '''Setup the test cases.

    Actions taken::

        - Create a temporary directory.

    '''
    import tempfile
    self.globs['outdir'] = tempfile.mkdtemp()


def _teardown(test):
    '''Cleanup the test cases.

    Actions taken::

        - Remove the temporary directory.

    '''
    import shutil
    shutil.rmtree(test.globs['outdir'])



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Render static-rest pages.
'''


from firmant import writers


class StaticRstWriter(writers.Writer):
    '''Render static-rest pages to the file system.

    :class:`StaticRstWriter` is an incomplete subclass of :class:`Writer`.  That
    is, there are still abstract methods/properties that are undefined.  In
    particular, `extension` and `render` must be defined.

    .. doctest::

       >>> class SampleStaticRstWriter(StaticRstWriter):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj):
       ...         print 'Save staticrst "%s" to "%s"' % (obj.path, path)
       ...
       >>> ssrw = SampleStaticRstWriter({}, {})
       >>> pprint(ssrw({}, {})) #doctest: +ELLIPSIS
       ({},
        {},
        [<firmant.writers.staticrst.SampleStaticRstWriter object at 0x...>,
         <firmant.writers.staticrst.SampleStaticRstWriter object at 0x...>])

    '''

    def key(self, staticrst):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a static-rest object:

            type
               This is always ``staticrst``.

            path
               A path that describes the object relative to the input/output
               directories.

        .. doctest::
           :hide:

           >>> class SampleStaticRstWriter(StaticRstWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save staticrst "%s" to "%s"' % (obj.path, path)
           ...
           >>> ssrw = SampleStaticRstWriter({}, {})

        .. doctest::

           >>> print objects.staticrst[0].path
           about
           >>> pprint(ssrw.key(objects.staticrst[0]))
           {'path': u'about', 'type': u'staticrst'}

        '''
        return {'type': u'staticrst', 'path': staticrst.path}

    def obj_list(self, environment, objects):
        '''Return all objects stored under the key ``staticrst``.

        .. doctest::
           :hide:

           >>> class SampleStaticRstWriter(StaticRstWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save staticrst "%s" to "%s"' % (obj.path, path)
           ...
           >>> ssrw = SampleStaticRstWriter({}, {})

        .. doctest::

           >>> ssrw.obj_list(None, {})
           []
           >>> ssrw.obj_list(None, {'staticrst': []})
           []
           >>> ssrw.obj_list(None, {'staticrst': ['staticobj']})
           ['staticobj']

        '''
        # pylint: disable-msg=W0613
        return objects.get('staticrst', [])


def _setup(test):
    '''Setup the environment for tests.
    '''
    from testdata.chunks import c900
    test.globs['objects'] = c900



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Setup Jinja2 globals (objects that should be available in Jinja2 templates).

Each Jinja2 writer will make the globals available in the context when rendering
templates.
'''


import datetime

from firmant import chunks
from firmant import decorators
from firmant.utils import workarounds
from firmant.writers import j2


class Global(chunks.AbstractChunk):
    '''Base class for other globals providers.

    Globals inject themselves into the environment[Jinja2Base]['globals']
    dictionary.

    :meth`__call__` will call :meth:`add_glob` with a dictionary, the
    environment and the objects.  The dictionary should be modified directly.
    '''

    def __init__(self, environment, objects):
        # pylint: disable-msg=W0613
        super(Global, self).__init__()

    def __call__(self, environment, objects):
        globs = self._get_globals_dict(environment)
        self.add_glob(globs, environment, objects)
        return environment, objects, []

    @workarounds.abstractmethod
    def add_glob(self, globs, environment, objects):
        '''Add the necessary globals to the dictionary `globs`.

        This dictionary will be copied for use in each Jinja2 context.
        '''

    @staticmethod
    def _get_globals_dict(environment):
        '''Return a the globals dictionary, creating it and the Jinja2
        environment if necessary.
        '''
        j2env = environment.get(j2.Jinja2Base, None)
        if j2env is None:
            environment[j2.Jinja2Base] = j2env = {}
        globs = j2env.get('globals', None)
        if globs is None:
            j2env['globals'] = globs = {}
        return globs

    scheduling_order = 800


class URLFor(Global):
    '''Add the function :func:`urlfor` to globals.

    The :func:`urlfor` is the :meth:`url` method of a
    :class:`firmant.routingURLMapper` object.
    '''

    @decorators.in_environment('urlmapper')
    def add_glob(self, globs, environment, objects):
        '''Create the ``urlfor`` global.
        '''
        globs['urlfor'] = environment['urlmapper'].url


def _sorted_posts(objects, newest_first=True):
    '''Return a list of sorted posts from objects
    '''
    return sorted(objects.get('post', []), reverse=newest_first,
                  key=lambda p: (p.published, p.slug))


class RecentPosts(Global):
    '''Add a list of the `SIDEBAR_POSTS_LEN` most recent posts to globals.
    '''

    @decorators.in_environment('settings')
    def add_glob(self, globs, environment, objects):
        '''Put the list of posts under the key ``recent_posts``.
        '''
        offset = environment['settings'].SIDEBAR_POSTS_LEN
        posts = _sorted_posts(objects)[:offset]
        globs['recent_posts'] = [(p.title, p.permalink) for p in posts]


class DailyArchives(Global):
    '''Add a list of the `SIDEBAR_POSTS_LEN` most recent daily archives to
    globals.
    '''

    @decorators.in_environment('settings')
    @decorators.in_environment('urlmapper')
    def add_glob(self, globs, environment, objects):
        '''Put the list of daily archives under the key ``daily_archives``.
        '''
        url = environment['urlmapper'].url
        offset = environment['settings'].SIDEBAR_POSTS_LEN
        posts = _sorted_posts(objects)
        archives = set([(p.published.year, p.published.month, p.published.day)
                        for p in posts])
        globs['daily_archives'] = [(datetime.date(y, m, d),
                                     url('html', type='post', year=y, month=m,
                                         day=d, page=1))
                                     for y, m, d in
                                     sorted(archives, reverse=True)][:offset]


class MonthlyArchives(Global):
    '''Add a list of the `SIDEBAR_POSTS_LEN` most recent monthly archives to
    globals.
    '''

    @decorators.in_environment('settings')
    @decorators.in_environment('urlmapper')
    def add_glob(self, globs, environment, objects):
        '''Put the list of monthly archives under the key ``monthly_archives``.
        '''
        url = environment['urlmapper'].url
        offset = environment['settings'].SIDEBAR_POSTS_LEN
        posts = _sorted_posts(objects)
        archives = set([(p.published.year, p.published.month) for p in posts])
        globs['monthly_archives'] = [(datetime.date(y, m, 1),
                                     url('html', type='post', year=y, month=m,
                                         page=1))
                                     for y, m in
                                     sorted(archives, reverse=True)][:offset]


class YearlyArchives(Global):
    '''Add a list of the `SIDEBAR_POSTS_LEN` most recent yearly archives to
    globals.
    '''

    @decorators.in_environment('settings')
    @decorators.in_environment('urlmapper')
    def add_glob(self, globs, environment, objects):
        '''Put the list of yearly archives under the key ``yearly_archives``.
        '''
        url = environment['urlmapper'].url
        offset = environment['settings'].SIDEBAR_POSTS_LEN
        posts = _sorted_posts(objects)
        archives = set([(p.published.year) for p in posts])
        globs['yearly_archives'] = [(datetime.date(y, 1, 1),
                                     url('html', type='post', year=y, page=1))
                                     for y in
                                     sorted(archives, reverse=True)][:offset]


class StaticPages(Global):
    '''Add a list of the static pages, sorted by title.
    '''

    def add_glob(self, globs, environment, objects):
        '''Put a list of all static pages under the key ``static_pages``.
        '''
        globs['static_pages'] = [(p.title, p.permalink) for p in
                sorted(objects.get('staticrst', []), key=lambda p: p.title)]


class AtomFeeds(Global):
    '''Add a list of the atom feeds, sorted by slug.
    '''

    def add_glob(self, globs, environment, objects):
        '''Put a list of all static pages under the key ``atom_feeds``.
        '''
        globs['atom_feeds'] = [(f.slug, f.permalink) for f in
                sorted(objects.get('feed', []) , key=lambda f: f.slug)]



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Copy static objects into the output directory.
'''


import os
import shutil

from firmant import writers


class StaticWriter(writers.Writer):
    '''Write copy or link the file pointed to by an object into the output
    directory.

    .. doctest::

       >>> sw = StaticWriter({}, {'static': objects.static})

    '''

    extension = None

    def key(self, static):
        '''Return the set of attributes suitable as input for url mapping.

        .. seealso::

           :meth:`firmant.parsers.static.StaticParser.attributes` for attributes
           that identify a static object.

        .. doctest::
           :hide:

           >>> sw = StaticWriter({}, {'static': objects.static})

        .. doctest::

           >>> print objects.static[0].relpath
           images/88x31.png
           >>> pprint(sw.key(objects.static[0]))
           {'path': 'images/88x31.png', 'type': u'static'}

        '''
        return {'type': u'static', 'path': static.relpath}

    def obj_list(self, environment, objects):
        '''Return all objects stored under the key ``static``.

        .. doctest::
           :hide:

           >>> sw = StaticWriter({}, {'static': objects.static})

        .. doctest::

           >>> sw.obj_list(None, {})
           []
           >>> sw.obj_list(None, {'static': []})
           []
           >>> sw.obj_list(None, {'static': ['staticobj']})
           ['staticobj']

        '''
        # pylint: disable-msg=W0613
        return objects.get('static', [])

    def render(self, environment, path, static):
        '''Copy the object to the correct place on the filesystem.
        '''
        try:
            os.makedirs(os.path.dirname(path))
        except OSError, ex:
            if ex.errno != 17:
                raise
        shutil.copy2(static.fullpath, path)


def _setup(test):
    '''Setup the environment for tests.
    '''
    from testdata.chunks import c900
    test.globs['objects'] = c900



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Static objects are simply references to files.

The most common use for static objects is copying images and css into the proper
location in the output hierarchy.
'''


import os

from firmant import decorators
from firmant import parsers


class StaticObject(parsers.ParsedObject):
    '''An object that serves as a placeholder for a file on the filesystem.

    Attributes of :class:`StaticObject`:

        fullpath
           A path that points to the object on the filesystem.  It is possible
           to `open` or copy using `fullpath`.

        relpath
           The path relative to the root where the static objects are stored.

    .. doctest::

       >>> StaticObject(fullpath='/tmp/path/to/static/object')
       static_obj</tmp/path/to/static/object>

    '''

    # pylint: disable-msg=R0903

    __slots__ = ['fullpath', 'relpath']

    def __repr__(self):
        return 'static_obj<%s>' % getattr(self, 'fullpath', None)

    @property
    def _attributes(self):
        '''Attributes that identify a static object:

            path
               A path that describes the object relative to the input/output
               directories.

        .. doctest::

           >>> pprint(StaticObject(relpath='images/88x31.png')._attributes)
           {'path': 'images/88x31.png'}

        '''
        # pylint: disable-msg=E1101
        return {'path': self.relpath}


class StaticParser(parsers.Parser):
    '''Create stand-in objects for static files to be published.

    .. doctest::
       :hide:

       >>> environment['log'] = get_logger()

    .. doctest::

       >>> sp = StaticParser(environment, objects)
       >>> environment, objects, (parser,) = sp(environment, objects)
       >>> pprint(parser(environment, objects)) #doctest: +ELLIPSIS
       ({...},
        {'static': [static_obj<testdata/pristine/static/images/88x31.png>]},
        [])

    '''

    type = 'static'
    paths = '.*'
    cls = StaticObject

    def parse(self, environment, objects, path):
        '''Parse the object at `path` and save it under ``objects[self.type]``
        '''
        fullpath = os.path.join(self.root(environment), path)
        objects[self.type].append(self.cls(fullpath=fullpath, relpath=path))

    @decorators.in_environment('settings')
    def root(self, environment):
        '''The directory under which all static objects reside.
        '''
        settings = environment['settings']
        return os.path.join(settings.CONTENT_ROOT, settings.STATIC_SUBDIR)


def _setup(test):
    '''Setup a test environment and test set of objects.
    '''
    from pysettings.settings import Settings
    test.globs['settings'] = Settings({'CONTENT_ROOT': 'testdata/pristine'
                                      ,'STATIC_SUBDIR': 'static'
                                      })
    test.globs['environment'] = {'settings': test.globs['settings']
                                }
    test.globs['objects'] = {}



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


'''Chunks drive the process of creating a webroot from an input directory.

Each chunk takes as arguments the settings, environment, and publishable
objects and returns a new environment, a new set of objects, and a set of
chunks that should be called in the future.

.. note::

   It must be the case that new chunks are scheduled such that they are called
   after the current chunk.

In this way, the flow of the application may be dynamically modified at the
higher level without placing many strict requirements on what is possible.  The
best practices that should be observed include:

 * A chunk should have a default policy of passing objects unchanged.  Only
   objects that the chunk is intended to modify should be modified by the
   chunks.
 * The same default-unchanged policy should apply to passing the non-object
   environment.
 * New chunks should have a scheduling order such that they are executed after
   the current chunk.

'''


import abc

from firmant.utils import workarounds


# pylint: disable-msg=R0903
class AbstractChunk(object):
    '''An :class:`AbstractChunk` defines the interface for all chunks.
    '''

    __metaclass__ = abc.ABCMeta

    @workarounds.abstractmethod
    def __call__(self, environment, objects):
        '''Execute the chunk with the state from environment/objects.

        The return value should be a tuple `(new_env, new_objs, new_chunks)`.
        The value of `new_env` will replace `environment` in calls to future
        chunks.  The value of `new_objs` will replace `objects` in future
        calls.  New chunks will be added to list of chunks to execute according
        to the scheduling order.

        '''

    @workarounds.abstractproperty
    def scheduling_order(self):
        ''':attr:`scheduling_order` is a positive integer that determines chunk
        execution order.

        :attr:`scheduling_order` uses normal integer comparison.  Chunks of the
        same priority will be executed in arbitrary order.

        '''



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Decorators used throughout Firmant.
'''


import inspect

import decorator


def in_environment(key):
    '''Assert that `key` is in the environment.

    .. doctest::

       >>> @in_environment('somekey')
       ... def testfunc(environment):
       ...     print 'testfunc executes'
       ...
       >>> inspect.getargspec(testfunc)
       ArgSpec(args=['environment'], varargs=None, keywords=None, defaults=None)
       >>> testfunc({'somekey': True})
       testfunc executes
       >>> testfunc(environment={'somekey': True})
       testfunc executes
       >>> testfunc({})
       Traceback (most recent call last):
       ValueError: Expected 'somekey' in 'environment'

    The decorated function must take a non-keyword attribute for the
    environment.

    .. doctest::

       >>> @in_environment('somekey')
       ... def testfunc():
       ...    print 'testfunc executes'
       ...
       Traceback (most recent call last):
       AttributeError: Decorated function must have arg 'environment'

    Environment cannot have a default value.

    .. doctest::

       >>> @in_environment('somekey')
       ... def testfunc(environment=None):
       ...    print 'testfunc executes'
       ...
       Traceback (most recent call last):
       AttributeError: Decorated function cannot have default value for 'environment'

    '''
    def internal(func):
        '''First internal wrapper.  This is the decorator itself (with key
        bound).
        '''
        _args, _varargs, _varkw, _defaults = inspect.getargspec(func)

        if 'environment' not in _args:
            error = _("Decorated function must have arg 'environment'")
            raise AttributeError(error)

        if _defaults is not None and \
           _args.index('environment') >= len(_args) - len(_defaults):
            error = _("Decorated function cannot have default value " +
                      "for 'environment'")
            raise AttributeError(error)

        def _wrapper_(_func, *args, **kwargs):
            '''The actual wrapped function.

            The signature is updated to match `func`
            '''
            if 'environment' in kwargs:
                environment = kwargs['environment']
            else:
                environment = args[_args.index('environment')]

            if key not in environment:
                error = _("Expected '%s' in 'environment'") % key
                raise ValueError(error)

            return _func(*args, **kwargs)
        return decorator.decorator(_wrapper_, func)
    return internal



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Functions for interacting with the docutils module.

In particular this module makes creation of directives much more straightforward
and simple.  The :func:`meta_data_directive` produces a
:class:`MetaDataDirective`.  This result is suitable for registering with
:func:`docutils.parsers.rst.directives.register_directive`.

Docutils' processing is a multi-step process in which a module is read, parsed,
transformed, and written.  Each directive created with
:func:`meta_data_directive` creates a :class:`MetaDataNode` in the document tree
created by docutils.  When the custom :class:`CustomTransformReader` class is
used as the reader for docutils, and is passed :func:`meta_data_transform`
class, these nodes that contain metadata are then removed from the document
tree, and the metadata they store is added to the dictionary that the
:func:`meta_data_transform` is closed around.

'''


import functools
import re

from docutils import io
from docutils import core
from docutils.parsers.rst import Directive
from docutils.parsers.rst import directives
from docutils.parsers.rst import roles
from docutils.transforms import Transform
from docutils.readers import standalone
from docutils import nodes

from firmant.utils import strptime


class MetaDataNode(nodes.Element, nodes.Invisible, nodes.Special):
    '''The MetaDataNode is a node that holds metadata for later Transforms.
    '''

    # pylint: disable-msg=R0904

    def __init__(self, details=None, rawsource='', *children, **attributes):
        nodes.Element.__init__(self, rawsource, *children, **attributes)

        self.details = details or dict()


class URLAttributeNode(nodes.Element, nodes.Invisible, nodes.Special):
    '''The URLAttributeNode is a node that url attributes for a later transform.
    '''

    # pylint: disable-msg=R0904

    def __init__(self, extension, urlattributes, urltext, rawsource='',
                 *children, **attributes):
        nodes.Element.__init__(self, rawsource, *children, **attributes)

        self.extension = extension
        self.attributes = urlattributes
        self.urltext = urltext


def generic_reference_role(regex, typ, convert, role, rawtext, text, lineno,
        inliner, options=None, content=None):
    '''A generic reference role.

    It is designed to be curried with a regex, the type of link, and a function
    to return the extension, attributes, and urltext from the groupdict of the
    regex.
    '''
    options = options or {}
    content = content or []
    match   = re.match(regex, text)
    if match is None:
        err = _('Improper format for `%s` reference.') % typ
        msg = inliner.reporter.error(err, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    attributes = match.groupdict()
    extension, attributes, urltext = convert(attributes)
    return [URLAttributeNode(extension, attributes, urltext)], []


_feed_reference_role_re = r'^(?P<extension>\w{0,6}): ' + \
                          r'(?P<slug>(?:\||\-|\w)+)\s(?P<text>.+)$'
def _feed_reference_role_convert(attributes):
    '''Pull the necessary data from the attributes.
    '''
    ret = attributes.copy()
    ret['type'] = 'feed'
    extension = ret['extension']
    del ret['extension']
    urltext = ret['text']
    del ret['text']
    return extension, ret, urltext
feed_reference_role = functools.partial(generic_reference_role,
        _feed_reference_role_re, 'feed', _feed_reference_role_convert)
roles.register_local_role('feed', feed_reference_role)


_post_reference_role_re = r'^(?P<extension>\w{0,6}): (?P<year>[0-9]{4})-' + \
                          r'(?P<month>[0-9]{2})-(?P<day>[0-9]{2})\s' + \
                          r'(?P<slug>(?:\||\-|\w)+)\s(?P<text>.+)$'
def _post_reference_role_convert(attributes):
    '''Pull the necessary data from the attributes.
    '''
    ret = attributes.copy()
    for val in ('year', 'month', 'day'):
        ret[val] = int(ret[val], 10)
    ret['type'] = 'post'
    extension = ret['extension']
    del ret['extension']
    urltext = ret['text']
    del ret['text']
    return extension, ret, urltext
post_reference_role = functools.partial(generic_reference_role,
        _post_reference_role_re, 'post', _post_reference_role_convert)
roles.register_local_role('post', post_reference_role)


_static_reference_role_re = r'(?P<path>[-_?%/a-zA-Z0-9.]+)\s(?P<text>.+)$'
def _static_reference_role_convert(attributes):
    '''Pull the necessary data from the attributes.
    '''
    ret = attributes.copy()
    ret['type'] = 'static'
    urltext = ret['text']
    del ret['text']
    return None, ret, urltext
static_reference_role = functools.partial(generic_reference_role,
        _static_reference_role_re, 'static',
        _static_reference_role_convert)
roles.register_local_role('static', static_reference_role)


_staticrst_reference_role_re = r'^(?P<extension>\w{0,6}): ' + \
                               r'(?P<path>[-_?%/a-zA-Z0-9]+)\s(?P<text>.+)$'
def _staticrst_reference_role_convert(attributes):
    '''Pull the necessary data from the attributes.
    '''
    ret = attributes.copy()
    ret['type'] = 'staticrst'
    extension = ret['extension']
    del ret['extension']
    urltext = ret['text']
    del ret['text']
    return extension, ret, urltext
staticrst_reference_role = functools.partial(generic_reference_role,
        _staticrst_reference_role_re, 'staticrst',
        _staticrst_reference_role_convert)
roles.register_local_role('staticrst', staticrst_reference_role)


_tag_reference_role_re = r'^(?P<extension>\w{0,6}): ' + \
                          r'(?P<slug>(?:\||\-|\w)+)\s(?P<text>.+)$'
def _tag_reference_role_convert(attributes):
    '''Pull the necessary data from the attributes.
    '''
    ret = attributes.copy()
    ret['type'] = 'tag'
    extension = ret['extension']
    del ret['extension']
    urltext = ret['text']
    del ret['text']
    return extension, ret, urltext
tag_reference_role = functools.partial(generic_reference_role,
        _tag_reference_role_re, 'tag', _tag_reference_role_convert)
roles.register_local_role('tag', tag_reference_role)


def meta_data_directive(func, whitespace=False):
    '''Create a Directive class to store data to a MetaDataNode.

    The function ``func`` is passed the contents of the node and a dictionary
    into which it should place all relevant metadata it extracts from the
    contents.
    '''

    class MetaDataDirective(Directive):
        '''A generic directive for capturing metadata.
        '''

        required_arguments = 0
        optional_arguments = 0
        final_argument_whitespace = whitespace
        option_spec = dict()
        has_content = True

        def run(self):
            '''This will be called by the docutils parsing process.

            It calls `func` with an empty dictionary, and the content that falls
            under the scope of the directive.  It then replaces said content
            with a :class:`MetaDataNode`.

            If func encounters an error (and wishes to cleanly display it to the
            user), it should throw a :exc:`ValueError` that should be presented
            using the `str` function.  The error itself will not be included in
            the final document.

            '''
            # Raise an error if the directive does not have contents.
            self.assert_has_content()

            # Parse the contents into metadata.
            d = dict()
            try:
                func(d, self.content)
            except ValueError, ex:
                self.state_machine.reporter.error(str(ex))
                return []

            # Insert a MetaDataNode with the metadata.
            node = MetaDataNode(d)
            return [node]

    return MetaDataDirective


def copyright_declaration(d, content):
    r'''Interpret the content as a copyright declaration.

    .. doctest::

       >>> d = dict()
       >>> copyright_declaration(d, [u'foo', u'bar'])
       >>> d['copyright']
       u'foo\nbar'

    '''
    d['copyright'] = '\n'.join(content)


def time(d, content):
    '''Interpret the content as a time.

    There are multiple formats that are accepted.

    .. todo::

       Make the formats accepted integrate with the localization routines.

    Using just hours:

    .. doctest::

       >>> d = dict()
       >>> time(d, ['15'])
       >>> d['time']
       datetime.time(15, 0)

    Using hours and minutes:

    .. doctest::

       >>> d = dict()
       >>> time(d, ['15:43'])
       >>> d['time']
       datetime.time(15, 43)

    Using hours, minutes, and seconds:

    .. doctest::

       >>> d = dict()
       >>> time(d, ['15:43:42'])
       >>> d['time']
       datetime.time(15, 43, 42)

    ValueError is raised on invalid time:

    .. doctest::

       >>> d = dict()
       >>> time(d, ['154342'])
       Traceback (most recent call last):
       ValueError: time data '154342' does not match any format.

    '''
    formats = ['%H', '%H:%M', '%H:%M:%S']
    dt = strptime(''.join(content), formats)
    d['time'] = dt.time()


def single_line(d, content, attr='line'):
    '''Interpret the content as a single line.

    A single line of content passed to this function will be stored as `line` in
    the dictionary `d`.

    .. doctest::

       >>> d = dict()
       >>> single_line(d, [u'foobar'])
       >>> d['line']
       u'foobar'

    Optionally, the attribute may be specified so that multiple `single_line`
    directives can co-exist.

    .. doctest::

       >>> d = dict()
       >>> single_line(d, [u'foobar'], 'attr')
       >>> d['attr']
       u'foobar'

    '''
    d[attr] = unicode(''.join(content))


def updated(d, content):
    '''Interpret content as a full datetime.

    There are multiple formats that are accepted.

    .. todo::

       Make the formats accepted integrate with the localization routines.

    YYYY-MM-DD HH:MM:SS (time may be omitted):

    .. doctest::

       >>> d = dict()
       >>> updated(d, ['2009-02-01 14:15:51'])
       >>> d['updated']
       datetime.datetime(2009, 2, 1, 14, 15, 51)

       >>> updated(d, ['2009-02-01'])
       >>> d['updated']
       datetime.datetime(2009, 2, 1, 0, 0)

    MM-DD-YYYY HH:MM:SS (time may be omitted):

    .. doctest::

       >>> d = dict()
       >>> updated(d, ['02-01-2009 14:15:51'])
       >>> d['updated']
       datetime.datetime(2009, 2, 1, 14, 15, 51)

       >>> updated(d, ['02-01-2009'])
       >>> d['updated']
       datetime.datetime(2009, 2, 1, 0, 0)

    ValueError is raised on invalid datetime:

    .. doctest::

       >>> d = dict()
       >>> updated(d, ['154342'])
       Traceback (most recent call last):
       ValueError: time data '154342' does not match any format.

    '''
    formats = ['%Y-%m-%d', '%Y-%m-%d %H', '%Y-%m-%d %H:%M',
               '%Y-%m-%d %H:%M:%S', '%m-%d-%Y', '%m-%d-%Y %H',
               '%m-%d-%Y %H:%M', '%m-%d-%Y %H:%M:%S']
    d['updated'] = strptime(''.join(content), formats)


def list_element(d, content, attr='element_plural'):
    '''Interpret content as an element in a list.

    It returns a list so that the metadata transform can append values to the
    list.

    .. doctest::

       >>> d = dict()
       >>> list_element(d, [u'foo'])
       >>> d['element_plural']
       [u'foo']

    The key in which to store the values may be specified.

    .. doctest::

       >>> d = dict()
       >>> list_element(d, [u'foo'], 'attr')
       >>> d['attr']
       [u'foo']

    '''
    d[attr] = list([''.join(content)])


directives.register_directive('copyright',
    meta_data_directive(copyright_declaration, whitespace=True))

directives.register_directive('time', meta_data_directive(time))

directives.register_directive('timezone',
    meta_data_directive(lambda d, c: single_line(d, c, 'timezone')))

directives.register_directive('author',
    meta_data_directive(lambda d, c: single_line(d, c, 'author')))

directives.register_directive('updated', meta_data_directive(updated))

directives.register_directive('tag',
    meta_data_directive(lambda d, c: list_element(d, c, 'tags')))

directives.register_directive('feed',
    meta_data_directive(lambda d, c: list_element(d, c, 'feeds')))


def meta_data_transform(data):
    '''Create a docutils Transform that is a closure around the `data` dict.
    '''

    # pylint: disable-msg=R0903

    class MetaDataTransform(Transform):
        '''Remove MetaDataNode nodes and set attributes on document.
        '''
        default_priority = 700

        def apply(self):
            '''This will be called by the docutils parsing process.

            The `data` dictionary from :class:`meta_data_transform` will be
            populated with the items in the `details` attribute of each
            MetaDataNode encountered.

            '''
            for node in self.document.traverse(MetaDataNode):
                for key, val in node.details.items():
                    try:
                        iter(val)
                        if key in data:
                            val = data[key] + val
                    except TypeError:
                        pass
                    data[key] = val
                node.parent.remove(node)

    return MetaDataTransform


def url_node_transform(urlmapper):
    '''Create a docutils Transform that is a closure around the `urlmapper`
    object.
    '''

    class URLNodeTransform(Transform):
        '''Transform :class:`URLAttributeNode` into docutils references.
        '''

        # pylint: disable-msg=R0903
        # pylint: disable-msg=R0913

        default_priority = 600

        def apply(self):
            '''This will be called by the docutils parsing process.
            '''
            for node in self.document.traverse(URLAttributeNode):
                refuri  = urlmapper.url(node.extension, **node.attributes)
                newnode = nodes.reference(node.urltext, node.urltext,
                                          refuri=refuri)
                node.replace_self(newnode)

    return URLNodeTransform


class CustomTransformsReader(standalone.Reader):
    '''Add transformations to read MetaData from doctree.
    '''

    def __init__(self, parser=None, parser_name=None, transforms=None):
        standalone.Reader.__init__(self, parser, parser_name)
        self.transforms = transforms or []

    def get_transforms(self):
        '''Add a transform for moving the meta data into the data dictionary.
        '''
        return standalone.Reader.get_transforms(self) + self.transforms


def publish(path, transforms=None):
    '''Publish the rst document that resides at `path` on the filesystem.

    This function returns a :class:`docutils.core.Publisher` object.  The
    optional `transforms` attribute may be passed with a list of transform
    classes that will be passed to the reader.

    '''
    args = {'source': None
           ,'source_path': path
           ,'source_class': io.FileInput
           ,'destination_class': io.StringOutput
           ,'destination': None
           ,'destination_path': None
           ,'reader': CustomTransformsReader(transforms=transforms or [])
           ,'reader_name': None
           ,'parser': None, 'parser_name': 'restructuredtext'
           ,'writer': None, 'writer_name': 'html'
           ,'settings': None, 'settings_spec': None
           ,'settings_overrides': None
           ,'config_section': None
           ,'enable_exit_status': None
           }

    # Code borrowed from :mod:`docutils.core` from version 0.5 of docutils.
    # This is the implementation of the :func:`publish_programmatically`
    # function.  It is reimplemented so that it may be fractured into
    # transformation steps later.

    # From the module:
    # Author: David Goodger <goodger@python.org>
    # Copyright: This module has been placed in the public domain.

    pub = core.Publisher(args['reader'], args['parser'], args['writer'],
            settings=args['settings'],
            source_class=args['source_class'],
            destination_class=args['destination_class'])
    pub.set_components(args['reader_name'], args['parser_name'],
            args['writer_name'])
    pub.process_programmatic_settings( args['settings_spec'],
            args['settings_overrides'], args['config_section'])
    pub.set_source(args['source'], args['source_path'])
    pub.set_destination(args['destination'], args['destination_path'])
    pub.publish(enable_exit_status=args['enable_exit_status'])

    # End borrowed code.

    return pub



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Workarounds for bugs/incompatibilities in upstream code.
'''


import abc


# Workaroudns for broken upstream :mod:`abc` http://bugs.python.org/issue8507
def abstractmethod(func):
    '''Replacement for broken upstream :func:`abc.abstractmethod`.
    '''
    _old_doc = func.__doc__
    func = abc.abstractmethod(func)
    # pylint: disable-msg=W0622
    func.__doc__ = _old_doc
    return func


def abstractproperty(prop):
    '''Replacement for broken upstream :func:`abc.abstractproperty`.
    '''
    _old_doc = prop.__doc__
    prop = abc.abstractproperty(prop)
    prop.__doc__ = _old_doc
    return prop



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Predefined path components for use with URLMapper.

Each of these pre-built components is meant for exactly one use case.  The
attributes (both free and bound) are shown, and then the various behaviors are
shown.  All should be self-explanatory.

TYPE
====

.. doctest::

   >>> TYPE('post').attributes
   set(['type'])
   >>> TYPE('post').bound_attributes
   {'type': 'post'}
   >>> TYPE('post').free_attributes
   set([])

   >>> TYPE('post').match(type='post')
   True
   >>> TYPE('post').match(type='tag')
   False

YEAR
====

.. doctest::

   >>> YEAR.attributes
   set(['year'])
   >>> YEAR.bound_attributes
   {}
   >>> YEAR.free_attributes
   set(['year'])

   >>> YEAR.construct(year=2010)
   '2010'
   >>> YEAR.construct(year=4)
   '0004'

MONTH
=====

.. doctest::

   >>> MONTH.attributes
   set(['month'])
   >>> MONTH.bound_attributes
   {}
   >>> MONTH.free_attributes
   set(['month'])

   >>> MONTH.construct(month=10)
   '10'
   >>> MONTH.construct(month=3)
   '03'

DAY
===

.. doctest::

   >>> DAY.attributes
   set(['day'])
   >>> DAY.bound_attributes
   {}
   >>> DAY.free_attributes
   set(['day'])

   >>> DAY.construct(day=15)
   '15'
   >>> DAY.construct(day=1)
   '01'

SLUG
====

.. doctest::

   >>> SLUG.attributes
   set(['slug'])
   >>> SLUG.bound_attributes
   {}
   >>> SLUG.free_attributes
   set(['slug'])

   >>> SLUG.construct(slug='foobar')
   'foobar'

PATH
====

.. doctest::

   >>> PATH.attributes
   set(['path'])
   >>> PATH.bound_attributes
   {}
   >>> PATH.free_attributes
   set(['path'])

   >>> PATH.construct(path='foo/bar')
   'foo/bar'

PAGENO
======

.. doctest::

   >>> PAGENO.attributes
   set(['page'])
   >>> PAGENO.bound_attributes
   {}
   >>> PAGENO.free_attributes
   set(['page'])

   >>> PAGENO.construct(page=1) is None
   True
   >>> PAGENO.construct(page=2)
   'page2'

'''


import functools

from firmant import routing


__all__ = ['TYPE', 'YEAR', 'MONTH', 'DAY', 'SLUG', 'PATH', 'PAGENO']


TYPE  = functools.partial(routing.BoundNullPathComponent, 'type')
YEAR  = routing.SinglePathComponent('year', lambda y: '%04i' % y)
MONTH = routing.SinglePathComponent('month', lambda m: '%02i' % m)
DAY   = routing.SinglePathComponent('day', lambda d: '%02i' % d)
SLUG  = routing.SinglePathComponent('slug', str)
PATH  = routing.SinglePathComponent('path', str)


def _page_conv(value):
    '''Conversion function for page numbers.
    '''
    if value == 1:
        return None
    return 'page%i' % value
PAGENO = routing.SinglePathComponent('page', _page_conv)



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Writers that inherit from other writers and render html.
'''


import jinja2

from firmant import decorators
from firmant.writers import staticrst
from firmant.writers import posts
from firmant.utils import paths
from firmant.utils import workarounds


class Jinja2Base(object):
    '''Base class used for functionality common to all J2 writers.
    '''

    # pylint: disable-msg=R0903

    @staticmethod
    @decorators.in_environment('settings')
    def render_to_file(environment, path, template, context):
        '''Render template with context and save to path.
        '''
        j2env = environment.get(Jinja2Base, {})
        environment[Jinja2Base] = j2env
        if 'env' not in j2env:
            settings = environment['settings']
            loader = getattr(settings, 'TEMPLATE_LOADER', None)
            if loader is None:
                loader = jinja2.PackageLoader('firmant', 'templates')
            j2env['env'] = jinja2.Environment(loader=loader)
        template = j2env['env'].get_template(template)
        globs    = j2env.get('globals', dict())
        globs.update(context)
        data     = template.render(globs)
        out      = paths.create_or_truncate(path)
        out.write(data.encode('utf-8'))
        out.flush()
        out.close()


class Jinja2StaticRst(Jinja2Base, staticrst.StaticRstWriter):
    '''Render staticrst objects using the :class:`StaticRstWriter` base.
    '''

    extension = 'html'

    template = 'flat.html'

    def render(self, environment, path, obj):
        '''Render the data in a Jinja2 template.

        .. doctest::

           >>> j2sr = Jinja2StaticRst(environment, objects)
           >>> obj = j2sr.obj_list(environment, objects)[0]
           >>> j2sr.render(environment, path, obj)
           >>> cat(path)
           About at about

        '''
        context = dict()
        context['path'] = obj.path
        context['page'] = obj
        self.render_to_file(environment, path, self.template, context)


class Jinja2PostWriter(Jinja2Base, posts.PostWriter):
    '''Render each post individually using Jinja2 templates.
    '''

    extension = 'html'

    template = 'posts/single.html'

    def render(self, environment, path, post):
        r'''Render the data in a Jinja2 template.

        .. doctest::

           >>> j2pw = Jinja2PostWriter(environment, objects)
           >>> obj = j2pw.obj_list(environment, objects)[0]
           >>> j2pw.render(environment, path, obj)
           >>> cat(path)
           2009-12-31 | party by John Doe

        '''
        context = dict()
        context['post']  = post
        self.render_to_file(environment, path, self.template, context)


class Jinja2PostArchiveBase(Jinja2Base):
    '''Common functionality for rendering Jinja2 archive views.
    '''

    # It complains about not having the key attribute (provided by children)
    # pylint: disable-msg=E1101

    extension = 'html'

    @workarounds.abstractproperty
    def template(self):
        '''The template to use for rendering.
        '''

    @decorators.in_environment('urlmapper')
    def render(self, environment, path, obj):
        '''Render the archive view.
        '''
        # pylint: disable-msg=R0912
        key = self.key(obj)
        context = dict()
        context['posts'] = obj[0]
        urlmapper = environment['urlmapper']
        if len(obj) == 3:
            pages = 2
            calen = 1
        elif len(obj) == 2:
            pages = 1
            calen = False
        # Construct the urls for prev/next paginated pages
        if obj[pages].prev:
            d = key.copy()
            d['page'] = obj[pages].prev
            context['page_prev'] = urlmapper.url(self.extension, **d)
        else:
            context['page_prev'] = None
        if obj[pages].next:
            d = key.copy()
            d['page'] = obj[pages].next
            context['page_next'] = urlmapper.url(self.extension, **d)
        else:
            context['page_next'] = None
        # Construct the urls for prev/next archive pages
        if calen and obj[calen].prev:
            d = key.copy()
            d['page'] = 1
            if len(obj[1].prev) >= 1:
                d['year'] = obj[calen].prev[0]
            if len(obj[1].prev) >= 2:
                d['month'] = obj[calen].prev[1]
            if len(obj[1].prev) >= 3:
                d['day'] = obj[calen].prev[2]
            context['cal_prev'] = urlmapper.url(self.extension, **d)
        else:
            context['cal_prev'] = None
        if calen and obj[calen].next:
            d = key.copy()
            d['page'] = 1
            if len(obj[1].next) >= 1:
                d['year'] = obj[calen].next[0]
            if len(obj[1].next) >= 2:
                d['month'] = obj[calen].next[1]
            if len(obj[1].next) >= 3:
                d['day'] = obj[calen].next[2]
            context['cal_next'] = urlmapper.url(self.extension, **d)
        else:
            context['cal_next'] = None
        self.render_to_file(environment, path, self.template, context)


class Jinja2PostArchiveAll(Jinja2PostArchiveBase, posts.PostArchiveAll):
    '''Render paginated post lists with Jinja2 templates.
    '''
    template = 'posts/archive_all.html'


class Jinja2PostArchiveYearly(Jinja2PostArchiveBase, posts.PostArchiveYearly):
    '''Render paginated post lists (grouped by year) with Jinja2 templates.
    '''
    template = 'posts/archive_yearly.html'


class Jinja2PostArchiveMonthly(Jinja2PostArchiveBase, posts.PostArchiveMonthly):
    '''Render paginated post lists (grouped by month) with Jinja2 templates.
    '''
    template = 'posts/archive_monthly.html'


class Jinja2PostArchiveDaily(Jinja2PostArchiveBase, posts.PostArchiveDaily):
    '''Render paginated post lists (grouped by day) with Jinja2 templates.
    '''
    template = 'posts/archive_daily.html'


def _setup(self):
    '''Setup the test cases.
    '''
    import tempfile
    import os
    from pysettings.settings import Settings
    from firmant.routing import URLMapper
    from firmant.utils.paths import cat
    from testdata.chunks import c900
    settings = Settings({'POSTS_PER_PAGE': 2
                        ,'OUTPUT_DIR': tempfile.mkdtemp()
                        ,'PERMALINK_ROOT': 'http://testurl'
                        ,'TEMPLATE_LOADER':
                        jinja2.FileSystemLoader('testdata/pristine/templates')
                        })
    urlmapper = URLMapper(settings.OUTPUT_DIR, settings.PERMALINK_ROOT)
    self.globs['settings']   = settings
    self.globs['cat']        = cat
    self.globs['environment'] = {'settings': settings
                                ,'urlmapper': urlmapper
                                ,Jinja2Base: {'globals':
                                              {'urlfor': urlmapper.url}}
                                }
    self.globs['objects'] = {'post': c900.posts, 'staticrst': c900.staticrst}
    self.globs['os'] = os
    self.globs['tempfile'] = os
    self.globs['path'] = tempfile.NamedTemporaryFile(delete=False).name
    self.globs['_path'] = self.globs['path']


def _teardown(test):
    '''Cleanup the Jinja2 test cases.
    '''
    import os
    import shutil
    os.unlink(test.globs['_path'])
    shutil.rmtree(test.globs['settings'].OUTPUT_DIR)



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Abstract base classes for writing posts.
'''


from firmant import decorators
from firmant import paginate
from firmant import writers


class PostWriter(writers.Writer):
    '''Write/render individual post objects.

    This post writer is serves as the basis for other post writers that write
    posts individually.

    .. doctest::

       >>> class SamplePost(PostWriter):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj):
       ...         print 'Save post "%s %s" to "%s"' % \
                       (obj.published.strftime('%Y-%m-%d'), obj.slug, path)
       >>> sp = SamplePost({}, {})
       >>> pprint(sp({}, {})) #doctest: +ELLIPSIS
       ({},
        {},
        [<firmant.writers.posts.SamplePost object at 0x...>,
         <firmant.writers.posts.SamplePost object at 0x...>])

    '''

    def key(self, post):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a post object:

            type
               This is always ``post``.

            year
               The year of publication.

            month
               The month of publication.

            day
               The day of publication.

            slug
               The slug that is unique to the feed object.

        .. doctest::
           :hide:

           >>> class SamplePost(PostWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save post "%s %s" to "%s"' % \
                           (obj.published.strftime('%Y-%m-%d'), obj.slug, path)
           >>> sp = SamplePost({}, {})

        .. doctest::

           >>> print objects.posts[0].published.date(), objects.posts[0].slug
           2009-12-31 party
           >>> pprint(sp.key(objects.posts[0]))
           {'day': 31, 'month': 12, 'slug': u'party', 'type': u'post', 'year': 2009}

        '''
        return {'type': u'post'
               ,'slug': post.slug
               ,'year': post.published.year
               ,'month': post.published.month
               ,'day': post.published.day
               }

    def obj_list(self, environment, objects):
        '''Return all objects stored under the key ``post``.

        .. doctest::
           :hide:

           >>> class SamplePost(PostWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save post "%s %s" to "%s"' % \
                           (obj.published.strftime('%Y-%m-%d'), obj.slug, path)
           >>> sp = SamplePost({}, {})

        .. doctest::

           >>> sp.obj_list(None, {})
           []
           >>> sp.obj_list(None, {'post': []})
           []
           >>> sp.obj_list(None, {'post': ['postobj']})
           ['postobj']

        '''
        return objects.get('post', [])


class PostArchiveAll(writers.Writer):
    '''Write the posts into paginated pages.

    .. doctest::

       >>> class SamplePostArchiveAll(PostArchiveAll):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spaa = SamplePostArchiveAll({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a archived year page:

            type
               This is always ``post``.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveAll(PostArchiveAll):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spaa = SamplePostArchiveAll({}, {})

        .. doctest::

           >>> obj = spaa.obj_list({'settings': settings},
           ...                     {'post': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([Post(2010-02-02-newday2), Post(2010-02-02-newday)], Paginated(None, 1, 2))

           >>> pprint(spaa.key(obj))
           {'page': 1, 'type': u'post'}

        '''
        return {'type': u'post'
               ,'page': obj[1].cur
               }

    @decorators.in_environment('settings')
    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by year, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveAll(PostArchiveAll):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spaa = SamplePostArchiveAll({}, {})

        .. doctest::

           >>> spaa.obj_list({'settings': settings}, {})
           []
           >>> spaa.obj_list({'settings': settings}, {'post': []})
           []
           >>> pprint(spaa.obj_list({'settings': settings},
           ...                      {'post': objects.posts})) #doctest: +ELLIPSIS
           [([Post(2010-02-02-newday2), Post(2010-02-02-newday)], Paginated(None, 1, 2)),
            ([Post(2010-02-01-newmonth), Post(2010-01-01-newyear)], Paginated(1, 2, 3)),
            ([Post(2009-12-31-party)], Paginated(2, 3, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('post', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.paginate(num_per_page, posts)


class PostArchiveYearly(writers.Writer):
    '''Write the posts into pages grouped by year and then paginated.

    .. doctest::

       >>> class SamplePostArchiveYearly(PostArchiveYearly):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spay = SamplePostArchiveYearly({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a archived year page:

            type
               This is always ``post``.

            year
               The year of publication.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveYearly(PostArchiveYearly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spay = SamplePostArchiveYearly({}, {})

        .. doctest::

           >>> obj = spay.obj_list({'settings': settings},
           ...                     {'post': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([Post(2010-02-02-newday2), Post(2010-02-02-newday)],
            Paginated(None, (2010,), (2009,)),
            Paginated(None, 1, 2))
           >>> pprint(spay.key(obj))
           {'page': 1, 'type': u'post', 'year': 2010}

        '''
        return {'type': u'post'
               ,'year': obj[1].cur[0]
               ,'page': obj[2].cur
               }

    @decorators.in_environment('settings')
    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by year, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveYearly(PostArchiveYearly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spay = SamplePostArchiveYearly({}, {})

        .. doctest::

           >>> spay.obj_list({'settings': settings}, {})
           []
           >>> spay.obj_list({'settings': settings}, {'post': []})
           []
           >>> pprint(spay.obj_list({'settings': settings},
           ...                      {'post': objects.posts}))
           [([Post(2010-02-02-newday2), Post(2010-02-02-newday)],
             Paginated(None, (2010,), (2009,)),
             Paginated(None, 1, 2)),
            ([Post(2010-02-01-newmonth), Post(2010-01-01-newyear)],
             Paginated(None, (2010,), (2009,)),
             Paginated(1, 2, None)),
            ([Post(2009-12-31-party)],
             Paginated((2010,), (2009,), None),
             Paginated(None, 1, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('post', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.split_paginate(num_per_page, self._splitfunc, posts)

    @staticmethod
    def _splitfunc(post):
        '''Split the posts at the year boundaries.
        '''
        return (post.published.year,)


class PostArchiveMonthly(writers.Writer):
    '''Write the posts into pages grouped by month and then paginated.

    .. doctest::

       >>> class SamplePostArchiveMonthly(PostArchiveMonthly):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spam = SamplePostArchiveMonthly({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a archived month page:

            type
               This is always ``post``.

            year
               The year of publication.

            month
               The month of publication.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveMonthly(PostArchiveMonthly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spam = SamplePostArchiveMonthly({}, {})

        .. doctest::

           >>> obj = spam.obj_list({'settings': settings},
           ...                     {'post': objects.posts})[0]
           >>> pprint(obj)
           ([Post(2010-02-02-newday2), Post(2010-02-02-newday)],
            Paginated(None, (2010, 2), (2010, 1)),
            Paginated(None, 1, 2))
           >>> pprint(spam.key(obj))
           {'month': 2, 'page': 1, 'type': u'post', 'year': 2010}

        '''
        return {'type': u'post'
               ,'year': obj[1].cur[0]
               ,'month': obj[1].cur[1]
               ,'page': obj[2].cur
               }

    @decorators.in_environment('settings')
    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by month, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveMonthly(PostArchiveMonthly):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spam = SamplePostArchiveMonthly({}, {})

        .. doctest::

           >>> spam.obj_list({'settings': settings}, {})
           []
           >>> spam.obj_list({'settings': settings}, {'post': []})
           []
           >>> pprint(spam.obj_list({'settings': settings},
           ...                      {'post': objects.posts})) #doctest: +ELLIPSIS
           [([Post(2010-02-02-newday2), Post(2010-02-02-newday)],
             Paginated(None, (2010, 2), (2010, 1)),
             Paginated(None, 1, 2)),
            ([Post(2010-02-01-newmonth)],
             Paginated(None, (2010, 2), (2010, 1)),
             Paginated(1, 2, None)),
            ([Post(2010-01-01-newyear)],
             Paginated((2010, 2), (2010, 1), (2009, 12)),
             Paginated(None, 1, None)),
            ([Post(2009-12-31-party)],
             Paginated((2010, 1), (2009, 12), None),
             Paginated(None, 1, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('post', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.split_paginate(num_per_page, self._splitfunc, posts)

    @staticmethod
    def _splitfunc(post):
        '''Split the posts at the month boundaries.
        '''
        return (post.published.year, post.published.month)


class PostArchiveDaily(writers.Writer):
    '''Write the posts into pages grouped by day and then paginated.

    .. doctest::

       >>> class SamplePostArchiveDaily(PostArchiveDaily):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj): pass
       >>> spad = SamplePostArchiveDaily({}, {})

    '''

    def key(self, obj):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify an archived day page:

            type
               This is always ``post``.

            year
               The year of publication.

            month
               The month of publication.

            day
               The day of publication.

            page
               The 1-indexed page number of the page.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveDaily(PostArchiveDaily):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spad = SamplePostArchiveDaily({}, {})

        .. doctest::

           >>> obj = spad.obj_list({'settings': settings},
           ...                     {'post': objects.posts})[0]
           >>> pprint(obj) #doctest: +ELLIPSIS
           ([Post(2010-02-02-newday2), Post(2010-02-02-newday)],
            Paginated(None, (2010, 2, 2), (2010, 2, 1)),
            Paginated(None, 1, None))
           >>> pprint(spad.key(obj))
           {'day': 2, 'month': 2, 'page': 1, 'type': u'post', 'year': 2010}

        '''
        return {'type': u'post'
               ,'year': obj[1].cur[0]
               ,'month': obj[1].cur[1]
               ,'day': obj[1].cur[2]
               ,'page': obj[2].cur
               }

    @decorators.in_environment('settings')
    def obj_list(self, environment, objects):
        '''Return 3-tuples with the list of objects and prev/next information.

        The prev/next information is stored in
        :class:`firmant.paginate.Paginated` objects and is returned for both the
        groupings by day, and for the groupings into pages.

        .. doctest::
           :hide:

           >>> class SamplePostArchiveDaily(PostArchiveDaily):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj): pass
           >>> spad = SamplePostArchiveDaily({}, {})

        .. doctest::

           >>> spad.obj_list({'settings': settings}, {})
           []
           >>> spad.obj_list({'settings': settings}, {'post': []})
           []
           >>> pprint(spad.obj_list({'settings': settings},
           ...                      {'post': objects.posts})) #doctest: +ELLIPSIS
           [([Post(2010-02-02-newday2), Post(2010-02-02-newday)],
             Paginated(None, (2010, 2, 2), (2010, 2, 1)),
             Paginated(None, 1, None)),
            ([Post(2010-02-01-newmonth)],
             Paginated((2010, 2, 2), (2010, 2, 1), (2010, 1, 1)),
             Paginated(None, 1, None)),
            ([Post(2010-01-01-newyear)],
             Paginated((2010, 2, 1), (2010, 1, 1), (2009, 12, 31)),
             Paginated(None, 1, None)),
            ([Post(2009-12-31-party)],
             Paginated((2010, 1, 1), (2009, 12, 31), None),
             Paginated(None, 1, None))]

        '''
        num_per_page = environment['settings'].POSTS_PER_PAGE
        posts = [p for p in reversed(sorted(objects.get('post', []),
                       key=lambda p: (p.published, p.slug)))]
        return paginate.split_paginate(num_per_page, self._splitfunc, posts)

    @staticmethod
    def _splitfunc(post):
        '''Split the posts at the day boundaries.
        '''
        return (post.published.year, post.published.month, post.published.day)


def _setup(test):
    '''Setup the environment for tests.
    '''
    from pysettings.settings import Settings
    from testdata.chunks import c900
    settings = Settings()
    settings.POSTS_PER_PAGE = 2
    test.globs['settings'] = settings
    test.globs['objects'] = c900



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Parsers create objects that can be manipulated and rendered into the final
webpage.

Modules in this package:

.. autosummary::
   :toctree: ../generated

   feeds
   posts
   static
   staticrst
   tags

'''


import abc
import logging
import stat
import tempfile
import os

from docutils import io
from docutils.core import publish_programmatically
from docutils.core import Publisher

from firmant import chunks
from firmant import decorators
from firmant import du
from firmant.utils import class_name
from firmant.utils import paths
from firmant.utils import workarounds
from firmant.utils.exceptions import log_uncaught_exceptions


class ParsedObject(object):
    '''A parsed object that represents the structures on disk in a form that is
    suitable for writing.

    The constructor will accept keyword arguments to automatically fill slots.

    .. doctest::

       >>> class SampleObject(ParsedObject):
       ...     __slots__ = ['someattr']
       ...     _attributes = property(lambda s: {'someattr': s.someattr})
       ...
       >>> SampleObject(someattr='value').someattr
       'value'
       >>> SampleObject(permalink='http://').permalink
       'http://'

    It is an error to provide arguments that are not in the slots of the class
    (or its base classes).

    .. doctest::

       >>> SampleObject(notinslots=True)
       Traceback (most recent call last):
       AttributeError: Excess attributes: 'notinslots'

    '''

    # pylint: disable-msg=R0903

    __metaclass__ = abc.ABCMeta

    __slots__ = ['permalink']

    def __init__(self, **kwargs):
        slots  = [set(getattr(cls, '__slots__', []))
                  for cls in self.__class__.__mro__]
        slots  = reduce(set.union, slots)
        excess = set(kwargs.keys()) - slots
        if len(excess) > 0:
            error = _("Excess attributes: '%s'") % \
                    "', '".join(excess)
            raise AttributeError(error)
        for attr in kwargs.keys():
            setattr(self, attr, kwargs.get(attr, None))

    @workarounds.abstractproperty
    def _attributes(self):
        '''The dict of attributes that define permalink of the object.

        The permalink will be derived from these attributes by passing them to
        a :class:`firmant.routing.URLMapper`.
        '''


class Parser(chunks.AbstractChunk):
    '''The base class of all chunks.

    This class defines an abstract base class that all parsers are required to
    adhere to.  To use this class in the creation of a parser, create a subclass
    with all necessary methods and properties overwritten.

    .. seealso::

       Module :mod:`abc`
          This module is part of the Python standard library in 2.6+.

    .. doctest::
       :hide:

       >>> import sys
       >>> logger = get_logger()

    To create a new type of parser, inherit from :class:`Parser`:

    .. doctest::

       >>> class SampleParser(Parser):
       ...     type = 'objs'
       ...     paths = '^numbers/[0-9]$'
       ...     cls = ParsedObject
       ...     def parse(self, environment, objects, path):
       ...         objects[self.type].append(str(path))
       ...     def attributes(self, environment, path):
       ...         return {'path': path}
       ...     def root(self, environment):
       ...          return 'testdata/pristine/sample'

    The new parser meets the criteria for two different abstract base classes:

    .. doctest::

       >>> import firmant.chunks
       >>> issubclass(SampleParser, firmant.chunks.AbstractChunk)
       True
       >>> issubclass(SampleParser, Parser)
       True

    .. warning::

       When creating a parser, do not store state in the parser itself.  While
       it appears that a parser is a single object, it will actually share state
       across two or more chunks during typical usage.

       If it is necessary to store state, place it in environment keyed to the
       parser class:

       .. doctest::
          :hide:

          >>> environment = {}

       .. doctest::

          >>> environment[SampleParser] = 'stored state goes here'

       This is because of the split between path selection and parsing.

    The remainder of this section is devoted to describing the implementation
    details of :class:`Parser`'s template methods.

    Chunks are passed environment and object dictionaries.  While it is not
    technically a chunk, the :class:`Parser` interface follows the same
    pattern.  When called with an environment and set of objects, a parser will
    return one more chunk (in addition to the environment and object
    dictionaries).

    .. doctest::

       >>> environment = {'log': logger
       ...               }
       >>> objects = {}
       >>> sp = SampleParser(environment, objects)
       >>> sp.scheduling_order
       10
       >>> pprint(sp(environment, objects)) #doctest: +ELLIPSIS
       ({'log': <logging.Logger instance at 0x...>},
        {},
        [<firmant.parsers.SampleParser object at 0x...>])

    .. note::

       The chunks returned do not share any state with the :class:`Parser` that
       created them.  The fact that the class name is the same is an
       implementation detail that may change in the future.

    The first chunk performs all parsing.  In the future, parsing may be broken
    into more fine-grained steps.  Right now, this is unnecessary.

    .. doctest::

       >>> environment, objects, (parse,) = sp(environment, objects)
       >>> parse.scheduling_order
       200
       >>> pprint(parse(environment, objects)) #doctest: +ELLIPSIS
       ({'log': <logging.Logger instance at 0x...>},
        {'objs': ['numbers/1', 'numbers/2', 'numbers/3']},
        [])

    ..
       These are for completeness of tests but don't contribute to docs.

    ..
       Make sure we don't accept bad actions.  This would break
       :attr:`scheduling_order` and :meth:`__call__`

    .. doctest::
       :hide:

       >>> SampleParser({}, {}, action='something else')
       Traceback (most recent call last):
       ValueError: `action` is invalid

    '''

    def __init__(self, environment, objects, action=None):
        # pylint: disable-msg=W0613
        super(Parser, self).__init__()
        if action not in (None, 'parse'):
            raise ValueError('`action` is invalid')
        self.__action__ = self.__default__
        if action == 'parse':
            self.__action__ = self.__parse__

    def __call__(self, environment, objects):
        return self.__action__(environment, objects)

    @property
    def scheduling_order(self):
        '''The following scheduling orders apply to parsers:

        10
           At timestep 10, the parser will create the chunks for finding paths
           and parsing.

        200
           Iterate over the paths in :meth:`paths` and pass each path to the
           :meth:`parse` method.  All new objects should be placed in the
           dictionary by :meth:`parse`.

        '''
        return {self.__default__: 10
               ,self.__parse__: 200}[self.__action__]

    def __default__(self, environment, objects):
        return (environment, objects,
                [self.__class__(environment, objects, 'parse')])

    @decorators.in_environment('log')
    def __parse__(self, environment, objects):
        for path in sorted(paths.recursive_listdir(self.root(environment),
            matches=self.paths)):
            if self.type not in objects:
                objects[self.type] = []
            environment['log'].info(_("%s parsing '%s'") %
                    (class_name(self.__class__), path))
            self.parse(environment, objects, path)
        return environment, objects, []

    @workarounds.abstractproperty
    def paths(self):
        '''Determine which paths should be parsed.

        This is a regular expression that will be used to match pathnames
        relative to :meth:`root`.
        '''

    @workarounds.abstractmethod
    def parse(self, environment, objects, path):
        '''Parse the object at path.

        Any new objects that are created during the parsing of the object at
        path should be added directly to the objects dictionary (this includes
        the parsed object itself).
        '''

    @workarounds.abstractproperty
    def type(self):
        '''The type of the primary object to be parsed (e.g. ``posts``).

        Parsed objects will be added to the objects dictionary under this key.

        This value has no impact on secondary objects that are generated (e.g.
        objects that are created from embedded LaTeX equations).
        '''

    @workarounds.abstractproperty
    def cls(self):
        '''The class object that should be used for new parsed objects.
        '''

    @workarounds.abstractmethod
    def root(self, environment):
        '''The root under which all objects to be parsed by this parser reside.
        '''


class RstMetaclass(abc.ABCMeta):
    '''A metaclass for dealing with :class:`RstParsedObject`.

    This class prvoides the magic that allows the document tree to be mutated
    and have the attributes of a :class:`RstParsedObject` update.

    '''

    # pylint: disable-msg=C0203

    def __init__(cls, name, bases, dct):
        super(RstMetaclass, cls).__init__(name, bases, dct)
        for attr, part in dct.get('_pubparts', []):
            @property
            def pubproperty(self, part=part):
                '''Property created by metaclass'''
                # pylint: disable-msg=W0212
                self._pub.writer.assemble_parts()
                return self._pub.writer.parts[part]
            setattr(cls, attr, pubproperty)


class RstParsedObject(ParsedObject):
    '''The base class of all objects parsed from reSt documents.
    '''

    # pylint: disable-msg=C0111
    # pylint: disable-msg=R0903

    __metaclass__ = RstMetaclass

    __slots__ = ['_pub', '_pubparts']


class RstParser(Parser):
    '''A parser containing common functionality for parsing reStructuredTest.
    '''

    @decorators.in_environment('urlmapper')
    def parse(self, environment, objects, path):
        '''Parse the reStructuredText doc at `path` and pass the relevant pieces
        to :meth:`rstparse`.
        '''
        metadata = {}
        transforms = [du.meta_data_transform(metadata),
                      du.url_node_transform(environment['urlmapper'])]
        pub = du.publish(os.path.join(self.root(environment), path), transforms)
        pieces = {}
        pieces['metadata'] = metadata
        pieces['document'] = pub.document
        pieces['pub'] = pub
        pieces['pub_parts'] = pub.writer.parts
        self.rstparse(environment, objects, path, pieces)

    @workarounds.abstractmethod
    def rstparse(self, environment, objects, path, pieces):
        '''The main method to override when creating new reStructuredText
        parsers.

        The `pieces` dictionary contains the following keys:

        metadata
           All pieces of metadata pulled from the document using
           :func:`firmant.du.meta_data_transform`.

        pub_parts
           The pieces returned by the html writer.

        document
           The actual doctree of that was produced as a result of parsing, and
           applying transformations.

        '''
        # TODO figure out how to handle transforms later (by deferring
        # attributes derived from pub_parts).



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''This module handles mapping between a set of attributes and a URL/document.

A :class:`URLMapper` serves as a means of converting a set of key-value pairs
into a full URL string.  Each instance of :class:`URLMapper` has a list of
objects that provide the :class:`AbstractPath` interface.  When querying
:meth:`URLMapper.url`, or :meth:`URLMapper.path`, the :class:`URLMapper` finds a
path object for which the provided attributes exactly cover the attributes of
the path.  Additionally, any bound variables must match the values provided in
the query.

.. note::

   Matching of attributes to URLs is designed to be similar to unification as
   found in Prolog, first order logic, and the lambda calculus; however, the
   similarity should not be read into too deeply as it may not be adhered to in
   the future.

Modules in this package:

.. autosummary::
   :toctree: ../generated

   components

'''


import abc
import os.path

from firmant.utils import class_name
from firmant.utils import merge_dicts
from firmant.utils import workarounds


__all__ = ['AbstractPath', 'BoundNullPathComponent', 'CompoundComponent',
        'SinglePathComponent', 'StaticPathComponent', 'URLMapper']


class AbstractPath(object):
    '''The interface for all path objects.

    The :class:`URLMapper` assumes all objects provided to it implement this
    interface.

    Intuitively, a path is simply a string that is defined by a set of
    key-value pairs.  The :attr:`attributes` property is the set of keys that
    define the path.

    The following properties must be true in any valid path object:

     * attributes = union(bound_attributes, free_attributes)
     * attributes - free_attributes = bound_attributes
     * attributes - bound_attributes = free_attributes

    '''

    __metaclass__ = abc.ABCMeta

    @workarounds.abstractproperty
    def attributes(self):
        '''All attributes that define the string representation of the path.
        '''

    @workarounds.abstractproperty
    def bound_attributes(self):
        '''Attributes with a fixed value.

        In order for a set of values to unify with this path, the keys and value
        of :attr:`bound_attributes` must match those specified in the query.

        '''

    @property
    def free_attributes(self):
        '''Attributes with no value.

        These attributes may be any value and still match the URL.

        '''
        return self.attributes - set(self.bound_attributes.keys())

    def match(self, *args, **kwargs):
        '''True if and only if `kwargs` and `args` specify the correct set of
        attributes.

        Each value in `args` should be a dictionary.  Provided keys must exactly
        cover all attributes and provided values must not conflict with bound
        attributes.  If the cover is not exact, or there are conflicting values,
        the result of :meth:`match` is False.

        '''
        # Do the merging of all values (if this throws a ValueError, we let it
        # rise because it is the fault of the caller).
        attrs = merge_dicts(kwargs, *args)

        # Check for conflict with bound values
        try:
            merge_dicts(attrs, self.bound_attributes)
        except ValueError:
            return False

        # Make sure the two sets are identical
        a = set(self.attributes)
        b = set(attrs.keys())
        return a & b == a | b

    @workarounds.abstractmethod
    def construct(self, *args, **kwargs):
        '''Use the values given in `kwargs` to construct the string
        representation of the path.

        The values in kwargs are not required to be any type, but should be of a
        type the path may safely convert to a string.

        '''

    def __div__(self, rhs):
        '''Concatenate two components (using :class:`CompoundComponent`).
        '''
        return CompoundComponent(self, rhs)


class SinglePathComponent(AbstractPath):
    '''A mapper between a single attribute and its string representation.

    Each instance will match exactly one attribute.  At creation time, a
    conversion function `conv` is specified that will be called to obtain the
    string representation of the attribute.

    In this example, a component with the attribute `month` is created.  When
    the :meth:`construct` method is called, it returns a two-digit
    representation of the number.

    .. doctest::

       >>> spc = SinglePathComponent('month', lambda m: '%02i' % m)
       >>> spc.attributes
       set(['month'])
       >>> spc.bound_attributes
       {}
       >>> spc.free_attributes
       set(['month'])
       >>> spc.match({'month': 3})
       True
       >>> spc.match(month=3)
       True
       >>> spc.match(year=2010)
       False
       >>> spc.match(year=2010, month=3)
       False
       >>> spc.construct({'month': 3})
       '03'
       >>> spc.construct(month=3)
       '03'

    It is an error to provide the same attribute twice with two different
    values:

    .. doctest::

       >>> spc.match({'month': 2}, month=3)
       Traceback (most recent call last):
       ValueError: Conflicting values for 'month'

    Note that the same attribute can be specified multiple times if the values
    are equal:

    .. doctest::

       >>> spc.match({'month': 3}, month=3)
       True

    If attributes not matching the single attribute are specified, then a
    :exc:`ValueError` will be thrown.

    .. doctest::

       >>> spc.construct(month=3, year=10)
       Traceback (most recent call last):
       ValueError: Attributes do not match path
       >>> spc.construct(year=10)
       Traceback (most recent call last):
       ValueError: Attributes do not match path
       >>> spc.construct()
       Traceback (most recent call last):
       ValueError: Attributes do not match path

    '''

    def __init__(self, attribute, conv=lambda x:x):
        super(SinglePathComponent, self).__init__()
        self._attribute = attribute
        self._conv = conv

    @property
    def attributes(self):
        '''The set of attributes contains the single attribute specified at
        creation time.
        '''
        return set([self._attribute])

    @property
    def bound_attributes(self):
        '''No attributes are bound.  This will always be an empty dictionary.
        '''
        return dict()

    def construct(self, *args, **kwargs):
        '''Create the string representation of the path.

        The value of the attribute will be passed through the conversion
        function.  The conversion function may return `None`.

        '''
        if not self.match(*args, **kwargs):
            raise ValueError('Attributes do not match path')
        attr = merge_dicts(kwargs, *args)
        return self._conv(attr[self._attribute])


class BoundNullPathComponent(AbstractPath):
    '''A mapper between a single attribute and the empty string.

    Each instance will match exactly one attribute.  Construct will always
    return `None` if there is a match.

    .. doctest::

       >>> bnpc = BoundNullPathComponent('month', 3)
       >>> bnpc.attributes
       set(['month'])
       >>> bnpc.bound_attributes
       {'month': 3}
       >>> bnpc.free_attributes
       set([])
       >>> bnpc.match({'month': 3})
       True
       >>> bnpc.match(month=3)
       True
       >>> bnpc.construct({'month': 3}) is None
       True
       >>> bnpc.construct(month=3) is None
       True

    It is an error to provide the same attribute twice with two different
    values:

    .. doctest::

       >>> bnpc.match({'month': 2}, month=3)
       Traceback (most recent call last):
       ValueError: Conflicting values for 'month'

    Note that the same attribute can be specified multiple times if the values
    are equal:

    .. doctest::

       >>> bnpc.match({'month': 3}, month=3)
       True

    If attributes not matching the single attribute are specified, then a
    :exc:`ValueError` will be thrown.

    .. doctest::

       >>> bnpc.construct(month=3, year=10)
       Traceback (most recent call last):
       ValueError: Attributes do not match path
       >>> bnpc.construct(year=10)
       Traceback (most recent call last):
       ValueError: Attributes do not match path
       >>> bnpc.construct()
       Traceback (most recent call last):
       ValueError: Attributes do not match path

    '''

    def __init__(self, attribute, value):
        super(BoundNullPathComponent, self).__init__()
        self._attribute = attribute
        self._value = value

    @property
    def attributes(self):
        '''The set of attributes contains the single attribute specified at
        creation time.

        '''
        return set([self._attribute])

    @property
    def bound_attributes(self):
        '''The single attribute is bound to the value specified.
        '''
        attrs = dict()
        attrs[self._attribute] = self._value
        return attrs

    def construct(self, *args, **kwargs):
        '''This will return `None` if the attributes match, or raise a
        :exc:`ValueError` otherwise.
        '''
        if not self.match(*args, **kwargs):
            raise ValueError('Attributes do not match path')
        return None


class StaticPathComponent(AbstractPath):
    '''A path component with no attributes that creates a static string.

    In this example, a path component is created that will construct to the path
    ``images``:

    .. doctest::

       >>> spc = StaticPathComponent('images')
       >>> spc.attributes
       set([])
       >>> spc.bound_attributes
       {}
       >>> spc.free_attributes
       set([])
       >>> spc.match({'month': 3})
       False
       >>> spc.match()
       True
       >>> spc.construct()
       'images'

    It is an error to specify any attributes when constructing:

    .. doctest::

       >>> spc.construct(month=2)
       Traceback (most recent call last):
       ValueError: Do not specify attributes for StaticPathComponent

    '''

    def __init__(self, path):
        super(StaticPathComponent, self).__init__()
        self._path = str(path)

    @property
    def attributes(self):
        '''By definition, a :class:`StaticPathComponent` has no attributes.

        This will return the empty set.

        '''
        return set([])

    @property
    def bound_attributes(self):
        '''By definition, a :class:`StaticPathComponent` has no attributes.

        This will return an empty dictionary.

        '''
        return {}

    def construct(self, *args, **kwargs):
        '''The constructed path is always the string constant specified.
        '''
        if not self.match(*args, **kwargs):
            raise ValueError(
                    'Do not specify attributes for StaticPathComponent')
        return self._path


class CompoundComponent(AbstractPath):
    '''A path object formed by joining two or more path objects with '/'.

    A :class:`CompoundComponent` is transparently formed by dividing path
    objects.

    .. doctest::

       >>> tags = StaticPathComponent('tags')
       >>> type = BoundNullPathComponent('type', 'tag')
       >>> slug = SinglePathComponent('slug', str)
       >>> path = type/tags/slug
       >>> path #doctest: +ELLIPSIS
       <firmant.routing.CompoundComponent object at 0x...>
       >>> path.match(type='tag', slug='foobar')
       True
       >>> path.match(type='tag')
       False
       >>> path.match(slug='foobar')
       False
       >>> path.construct(type='tag', slug='foobar')
       'tags/foobar'

    If :meth:`construct` is called with attributes that do not match the path,
    then a :exc:`ValueError` will be thrown.

    .. doctest::

       >>> path.construct(type='tag')
       Traceback (most recent call last):
       ValueError: Attributes do not match path

    '''

    def __init__(self, *args):
        super(CompoundComponent, self).__init__()
        self._components = args
        bound_attributes = [c.bound_attributes for c in self._components]
        self._bound = merge_dicts({}, *bound_attributes)

    @property
    def attributes(self):
        '''The attributes are the union of all components' attributes.
        '''
        return reduce(set.__or__, [c.attributes for c in self._components])

    @property
    def bound_attributes(self):
        '''The attributes are the union of all components' bound_attributes.
        '''
        return self._bound

    def construct(self, *args, **kwargs):
        '''Construct each path component, and join with '/'.
        '''
        if not self.match(*args, **kwargs):
            raise ValueError('Attributes do not match path')
        bound = merge_dicts(kwargs, *args)
        list_of_paths = [self._construct(x, bound) for x in self._components]
        list_of_paths = [p for p in list_of_paths if p]
        return '/'.join(list_of_paths)

    @classmethod
    def _construct(cls, component, d):
        '''A utility function for construct.

        Keys in `d` that are not in `component.attributes` are discarded before
        calling `component.construct`.

        '''
        d = d.copy()
        for key in d.keys():
            if key not in component.attributes:
                d.pop(key, None)
        return component.construct(d)


class URLMapper(object):
    '''Find the url or filesystem path that correlate with a set of attributes.

    The distinction between urls and paths is best described by example.  Let's
    declare the attributes ``slug='foo'`` and ``type='object'``.  The path on
    the filesystem, where the output would be written could be
    ``/path/to/output/directory/objects/foo/index.html`` while the URL where the
    document is accessible would be ``http://permanent.url/objects/foo/``.

    Having the URLMapper handle the logic of both paths and URLs makes sense.
    Consider a case where the user wishes to have the above URL be
    ``http://permanent.url/objects/foo.html``.  The local filesystem path
    would need to adapt to ``/path/to/output/directory/objects/foo.html``

    Creating a URLMapper is simple:

    .. doctest::

       >>> from firmant.routing.components import *
       >>> um = URLMapper('/path/to/output/directory', 'http://permanent.url/')
       >>> um.add( TYPE('post')/YEAR/MONTH/DAY/SLUG )
       >>> um.add( TYPE('post')/YEAR/MONTH/DAY )
       >>> um.add( TYPE('post')/YEAR/MONTH )
       >>> um.add( TYPE('post')/YEAR )

    Mapping a set of attributes to a path or URL is a matter of specifying the
    extension of the document (e.g. 'html' or 'css') and a set of key-value
    attributes.

    .. doctest::

       >>> um.path('html', type='post', slug='foo', day=15, month=3, year=2010)
       '/path/to/output/directory/2010/03/15/foo/index.html'
       >>> um.url('html', type='post', slug='foo', day=15, month=3, year=2010)
       'http://permanent.url/2010/03/15/foo/'

    If the attributes do not correspond to any path definition, then the value
    `None` is returned:

    .. doctest::

       >>> um.path('html', non_existent_attribute=True)
       >>> um.url('html', non_existent_attribute=True)

    If the extension is `None`, then the :meth:`path` and :meth:`url` methods
    will not add any information to account for an extension.

    .. doctest::

       >>> um.path(None, type='post', slug='foo', day=15, month=3, year=2010)
       '/path/to/output/directory/2010/03/15/foo'
       >>> um.url(None, type='post', slug='foo', day=15, month=3, year=2010)
       'http://permanent.url/2010/03/15/foo'

    As a special corner case, the :meth:`url` method will return the permalink
    root if it is asked for ``extension=None`` with no arguments.

    .. doctest::

       >>> um.url(None)
       'http://permanent.url/'

    This is useful when it is known that the attributes specified promise to
    resolve to a path.  Example uses include static files that are simply copied
    into the output directory.

    '''

    def __init__(self, output_dir, url_root, urls=None):
        self.output_dir = output_dir
        self.url_root   = url_root

        self._paths = list()
        if urls is not None:
            self._paths.extend(urls)

    def add(self, path):
        '''Add the path to the list of paths in the :class:`URLMapper`
        '''
        self._paths.append(path)

    def __lookup__(self, **kwargs):
        for path in self._paths:
            if path.match(kwargs):
                return path.construct(kwargs)
        raise AttributeError('Attributes do not correspond to any path')

    def url(self, extension, **kwargs):
        '''Return the URL corresponding to a set of attributes.
        '''
        if extension is None and kwargs == {}:
            path = ''
        else:
            try:
                path = self.__lookup__(**kwargs)
            except AttributeError:
                return None
        path = os.path.join(self.url_root, path or '')
        if extension is not None and not path.endswith('/'):
            path += '/'
        return path

    def path(self, extension, **kwargs):
        '''Return the filesystem path corresponding to a set of attributes.
        '''
        try:
            path = self.__lookup__(**kwargs)
        except AttributeError:
            return None
        path = os.path.join(self.output_dir, path or '')
        if extension is not None:
            path = os.path.join(path, 'index.%s' % extension)
        return path



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Writers transform parsed objects into files and directories on disk.

Writers are implemented behind the scenes in terms of chunks as defined in
:mod:`firmant.chunks`.  This is because writers need to do work at two different
points in time, with other chunks being scheduled in between.  The first chunk
executed by a writer will forward-declare URLs that the writer claims
responsibility for.  The second chunk renders objects and saves the result to
disk.

Modules in this package:

.. autosummary::
   :toctree: ../generated

   atom
   feeds
   j2
   posts
   static
   staticrst

'''


from firmant import chunks
from firmant import decorators
from firmant import utils
from firmant.utils import workarounds


class Writer(chunks.AbstractChunk):
    '''The base class of all writers.

    This class defines an abstract base class that all writers are required to
    adhere to.  To use this class in the creation of a writer, create a subclass
    with all necessary methods and properties overwritten.

    .. seealso::

       Module :mod:`abc`
          This module is part of the Python standard library in 2.6+.

    .. doctest::
       :hide:

       >>> import sys
       >>> from firmant import routing
       >>> logger = get_logger()

    To create a new type of writer, inherit from :class:`Writer`: 

    .. doctest::

       >>> class SampleWriter(Writer):
       ...     extension = 'txt'
       ...     def key(self, obj):
       ...         return {'obj': str(obj)}
       ...     def obj_list(self, environment, objects):
       ...         return objects.get('objs', [])
       ...     def render(self, environment, path, obj):
       ...         print 'Save object "%s" to "%s"' % (obj, path)

    The new writer meets the criteria for two different abstract base classes:

    .. doctest::

       >>> import firmant.chunks
       >>> issubclass(SampleWriter, firmant.chunks.AbstractChunk)
       True
       >>> issubclass(SampleWriter, Writer)
       True

    .. warning::

       When creating a writer, do not store state in the writer itself.  While
       it appears that a writer is a single object, it will actually share state
       across three or more chunks during typical usage.

       If it is necessary to store state, place it in environment keyed to the
       writer class:

       .. doctest::
          :hide:

          >>> environment = {}

       .. doctest::

          >>> environment[SampleWriter] = 'stored state goes here'

       This is because of the need to split actions between url
       forward-declaration and rendering.

    The remainder of this section is devoted to describing the implementation
    details of :class:`Writer`'s template methods.

    Chunks are passed environment and object dictionaries.  While it is not
    technically a chunk, the :class:`Writer` interface follows the same pattern.
    When called with an environment and set of objects, a writer will return two
    more chunks (in addition to the environment and object dictionaries).

    .. doctest::

       >>> environment = {'log': logger
       ...               ,'urlmapper': urlmapper
       ...               }
       >>> environment['urlmapper'].add(
       ...     routing.SinglePathComponent('obj', str)
       ... )
       >>> objects = {'objs': ['obj1', 'obj2', 'obj3']} 
       >>> sw = SampleWriter(environment, objects)
       >>> sw.scheduling_order
       10
       >>> pprint(sw(environment, objects)) #doctest: +ELLIPSIS
       ({'log': <logging.Logger instance at 0x...>,
         'urlmapper': <firmant.routing.URLMapper object at 0x...>},
        {'objs': ['obj1', 'obj2', 'obj3']},
        [<firmant.writers.SampleWriter object at 0x...>,
         <firmant.writers.SampleWriter object at 0x...>])

    .. note::

       The chunks returned do not share any state with the :class:`Writer` that
       created them.  The fact that the class name is the same is an
       implementation detail that may change in the future.

    The first chunk is the chunk that will build the urls, while the second is
    used for rendering.  Neither chunk returns more chunks.

    .. doctest::

       >>> environment, objects, (urls, render) = sw(environment, objects)
       >>> urls.scheduling_order
       500
       >>> render.scheduling_order
       900
       >>> pprint(urls(environment, objects)) #doctest: +ELLIPSIS
       ({'log': <logging.Logger instance at 0x...>,
         'urlmapper': <firmant.routing.URLMapper object at 0x...>,
         'urls': {'firmant.writers.SampleWriter': ['http://testurl/obj1/',
                                                   'http://testurl/obj2/',
                                                   'http://testurl/obj3/']}},
        {'objs': ['obj1', 'obj2', 'obj3']},
        [])
       >>> pprint(render(environment, objects)) #doctest: +ELLIPSIS
       Save object "obj1" to "outputdir/obj1/index.txt"
       Save object "obj2" to "outputdir/obj2/index.txt"
       Save object "obj3" to "outputdir/obj3/index.txt"
       ({'log': <logging.Logger instance at 0x...>,
         'urlmapper': <firmant.routing.URLMapper object at 0x...>},
        {'objs': ['obj1', 'obj2', 'obj3']},
        [])

    ..
       These are for completeness of tests but don't contribute to docs.

    ..
       Make sure we don't accept bad actions.  This would break
       :attr:`scheduling_order` and :meth:`__call__`

    .. doctest::
       :hide:

       >>> SampleWriter({}, {}, action='something else')
       Traceback (most recent call last):
       ValueError: `action` is invalid

    '''

    def __init__(self, environment, objects, action=None):
        # pylint: disable-msg=W0613
        super(Writer, self).__init__()
        if action not in (None, 'urls', 'renderer'):
            raise ValueError('`action` is invalid')
        self.__action__ = self.__default__
        if action == 'urls':
            self.__action__ = self.__urls__
        elif action == 'renderer':
            self.__action__ = self.__renderer__

    def __call__(self, environment, objects):
        return self.__action__(environment, objects)

    @property
    def scheduling_order(self):
        '''The following scheduling orders apply to writers:

        10
           At timestep 10, the writer will create the chunks for forward
           declaration and rendering.

        500
           At timestep 500, the writer will forward-declare urls.

        900
           At timestep 900, the writer will render the objects.
        '''
        return {self.__default__: 10
               ,self.__urls__: 500
               ,self.__renderer__: 900}[self.__action__]

    def __default__(self, environment, objects):
        return (environment, objects,
                [self.__class__(environment, objects, 'urls'),
                 self.__class__(environment, objects, 'renderer')])

    @decorators.in_environment('log')
    @decorators.in_environment('urlmapper')
    def __urls__(self, environment, objects):
        urlmapper = environment['urlmapper']
        newenv = environment.copy()
        if 'urls' not in newenv:
            newenv['urls'] = {}
        newenv['urls'][self.writername] = ret = []
        for obj in self.obj_list(environment, objects):
            url = urlmapper.url(self.extension, **self.key(obj))
            environment['log'].info(_("%s declared '%s'") %
                    (utils.class_name(self.__class__), url))
            ret.append(url)
        ret.sort()
        return (newenv, objects, [])

    @decorators.in_environment('urlmapper')
    def __renderer__(self, environment, objects):
        if 'urlmapper' not in environment:
            error = _('`urlmapper` expected in `environment`')
            raise ValueError(error)
        urlmapper = environment['urlmapper']
        for obj in self.obj_list(environment, objects):
            path = urlmapper.path(self.extension, **self.key(obj))
            environment['log'].info(_("%s rendered '%s'") %
                    (utils.class_name(self.__class__), path))
            self.render(environment, path, obj)
        return (environment, objects, [])

    @workarounds.abstractproperty
    def extension(self):
        '''The extension that will be used when finding the path/url of an
        object.

        This will be passed to a :class:`firmant.routing.URLMapper` instance.
        '''

    @workarounds.abstractmethod
    def key(self, obj):
        '''Map the object to a dictionary of attributes.

        The attributes will be passed to the :meth:`path` and :meth`url` methods
        of :class:`firmant.routing.URLMapper`.
        '''

    @workarounds.abstractmethod
    def obj_list(self, environment, objects):
        '''The objects that should be passed to :meth:`render`

        It will be passed the `environment` and `objects` dictionaries that were
        passed to the chunk.
        '''

    @workarounds.abstractmethod
    def render(self, environment, path, obj):
        '''Write the object to the path on filesystem.

        `path` will be a path under the output directory.  `obj` is one of
        the objects returned by obj_list.
        '''

    @property
    def writername(self):
        '''The string displayed when interacting with the user.

        This should be the name of the class they must specify in the
        configuration.  The default value should not be changed.
        '''
        return utils.class_name(self.__class__)


def _setup(self):
    '''Setup the test cases.
    '''
    from firmant.routing import URLMapper
    self.globs['urlmapper'] = URLMapper('outputdir', 'http://testurl/')



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Abstract base classes for writing feeds.
'''


from firmant import writers


class FeedWriter(writers.Writer):
    '''Write/render individual feed objects.

    This feed writer is serves as the basis for other feed writers (e.g. Atom or
    RSS).

    .. doctest::

       >>> class SampleFeed(FeedWriter):
       ...     extension = 'txt'
       ...     def render(self, environment, path, obj):
       ...         print 'Save feed "%s" to "%s"' % (obj.slug, path)
       ...
       >>> sf = SampleFeed({}, {})
       >>> pprint(sf({}, {})) #doctest: +ELLIPSIS
       ({},
        {},
        [<firmant.writers.feeds.SampleFeed object at 0x...>,
         <firmant.writers.feeds.SampleFeed object at 0x...>])

    '''

    def key(self, feed):
        '''Return the set of attributes suitable as input for url mapping.

        Attributes that identify a feed object:

            type
               This is always ``feed``.

            slug
               The slug that is unique to the feed object.

        .. doctest::
           :hide:

           >>> class SampleFeed(FeedWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save feed "%s" to "%s"' % (obj.slug, path)
           ...
           >>> sf = SampleFeed({}, {'feed': objects.feeds})

        .. doctest::

           >>> print objects.feeds[0].slug
           foo
           >>> pprint(sf.key(objects.feeds[0]))
           {'slug': u'foo', 'type': u'feed'}

        '''
        return {'type': u'feed', 'slug': feed.slug}

    def obj_list(self, environment, objects):
        '''Return all objects stored under the key ``feed``.

        .. doctest::
           :hide:

           >>> class SampleFeed(FeedWriter):
           ...     extension = 'txt'
           ...     def render(self, environment, path, obj):
           ...         print 'Save feed "%s" to "%s"' % (obj.slug, path)
           ...
           >>> sf = SampleFeed({}, {'feed': objects.feeds})

        .. doctest::

           >>> sf.obj_list(None, {})
           []
           >>> sf.obj_list(None, {'feed': []})
           []
           >>> sf.obj_list(None, {'feed': ['feedobj']})
           ['feedobj']

        '''
        return objects.get('feed', [])


def _setup(test):
    '''Setup the environment for tests.
    '''
    from testdata.chunks import c900
    test.globs['objects'] = c900



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Render feed in the `Atom format
<http://www.atomenabled.org/developers/syndication/atom-format-spec.php>`_.

'''


from lxml import etree

from firmant.utils import paths
from firmant.writers.feeds import FeedWriter


def add_text_subelement(root, name, text):
    '''Add a subelement named `name` to `root` with `text` inside.
    '''
    new = etree.SubElement(root, name)
    new.text = text


def rfc3339(dt):
    '''Render the datetime object `dt` properly accoring to rfc 3339.

    It works properly without the time zone.

    .. doctest::

       >>> dt = datetime.datetime(2010, 05, 02, 01, 02, 03)
       >>> rfc3339(dt)
       '2010-05-02T01:02:03Z'

    It works properly with the time zone.

    .. doctest::

       >>> import pytz
       >>> eastern = pytz.timezone('US/Eastern')
       >>> dt = eastern.localize(datetime.datetime(2010, 05, 02, 01, 02, 03))
       >>> rfc3339(dt)
       '2010-05-02T01:02:03-04:00'

    '''
    frmt_str = '%Y-%m-%dT%H:%M:%S'
    timestamp = dt.strftime(frmt_str)
    timezone  = dt.strftime('%z')
    if timezone != '':
        return timestamp + timezone[:-2] + ':' + timezone[-2:]
    else:
        return timestamp + 'Z'


class AtomFeed(FeedWriter):
    '''Write feeds in the Atom format.

    Uses :class:`firmant.writers.feeds.FeedWriter` to render feeds in Atom
    format with the extension ``atom``.

    .. doctest::

       >>> AtomFeed({}, {'feed': objects.feeds}) #doctest: +ELLIPSIS
       <firmant.writers.atom.AtomFeed object at 0x...>

    '''

    extension = 'atom'

    def render(self, environment, path, feed_obj):
        '''Render the feed according to the Atom specification.
        '''
        feed = etree.Element('feed')
        feed.set('xmlns', 'http://www.w3.org/2005/Atom')

        add_text_subelement(feed, 'generator', 'Firmant')
        add_text_subelement(feed, 'title', feed_obj.title)
        add_text_subelement(feed, 'rights', feed_obj.copyright)
        add_text_subelement(feed, 'updated', rfc3339(feed_obj.updated))
        add_text_subelement(feed, 'id', feed_obj.permalink)
        link = etree.SubElement(feed, 'link')
        link.set('href', feed_obj.permalink)
        link.set('rel', 'self')

        for post_obj in feed_obj.posts:
            post = etree.SubElement(feed, 'entry')
            add_text_subelement(post, 'title', post_obj.title)
            add_text_subelement(post, 'updated', rfc3339(post_obj.updated))
            add_text_subelement(post, 'published', rfc3339(post_obj.published))

            author = etree.SubElement(post, 'author')
            add_text_subelement(author, 'name', post_obj.author)

            content      = etree.SubElement(post, 'content')
            content.text = post_obj.content
            content.set('type', 'html')

            l_alt = etree.SubElement(post, 'link')
            l_alt.set('href', post_obj.permalink)
            l_alt.set('rel', 'alternate')

            add_text_subelement(post, 'id', post_obj.permalink)
            add_text_subelement(post, 'rights', post_obj.copyright)

            for tag in post_obj.tags:
                category = etree.SubElement(post, 'category')
                category.set('term', tag.slug)
                category.set('label', tag.title)

        out    = paths.create_or_truncate(path)
        out.write(etree.tostring(feed).encode('utf-8'))
        out.flush()
        out.close()


def _setup(test):
    '''Setup the environment for tests.
    '''
    import datetime
    from testdata.chunks import c900
    test.globs['objects'] = c900
    test.globs['datetime'] = datetime



from distutils.core import setup


classifiers = [ 'Development Status :: 4 - Beta'
              , 'Intended Audience :: Developers'
              , 'License :: OSI Approved :: BSD License'
              , 'Operating System :: MacOS :: MacOS X'
              , 'Operating System :: POSIX :: Linux'
              , 'Operating System :: Unix'
              , 'Programming Language :: Python :: 2.6'
              , 'Topic :: Internet :: WWW/HTTP :: Site Management'
              ]


setup(name='Firmant',
      version='0.2.1',
      author='Robert Escriva (rescrv)',
      author_email='firmant@mail.robescriva.com',
      packages=['firmant'
               ,'firmant.parsers'
               ,'firmant.routing'
               ,'firmant.templates'
               ,'firmant.utils'
               ,'firmant.writers'
               ],
      package_dir={'firmant': 'firmant'},
      package_data={'firmant': ['templates/*.html',
                                'templates/*/*.html']},
      scripts=['bin/firmant'],
      url='http://firmant.org/',
      license='3-clause BSD',
      description='A framework for static web applications.',
      long_description=open('doc/README.rst').read(),
      classifiers=classifiers,
      )



# Copyright (c) 2010, Robert Escriva
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Firmant nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''Manipulate feed objects for Firmant.
'''


import datetime
import os

import pytz

from firmant import decorators
from firmant import parsers


class Feed(parsers.RstParsedObject):
    '''A feed is a means of categorizing objects.

    Attributes of :class:`Feed`:

        slug
           A unique string that identifies the feed.

        title
           A longer title used for identifying the feed to the user.

        subtitle
           An even longer description of the feed that may be displayed to the
           user in combination with `title`.

        copyright
           The copyright of the feed for all posts without an explicit
           copyright.

        content
           The content of the restructured text document.  This does not include
           the title information.

        posts
           A list of cross-referenced posts.  This will be blank until
           cross-referencing happens at a later point in time.

    .. note::

       All string attributes except `slug` may be ``''``.  Posts will be ``[]``
       until the cross-referencing chunk happens.

    .. doctest::

       >>> Feed(slug='foo')
       Feed(foo)

    '''

    # pylint: disable-msg=R0903

    __slots__ = ['slug', 'copyright', 'posts']

    _pubparts = [('content', 'fragment')
                   ,('title', 'title')
                   ,('subtitle', 'subtitle')
                   ]

    def __repr__(self):
        return 'Feed(%s)' % getattr(self, 'slug', None)

    @property
    def _attributes(self):
        '''Attributes that identify a feed object:

            slug
               The `slug` attribute of the :class:`Feed` object.

        .. doctest::

           >>> Feed(slug='foo')._attributes
           {'slug': 'foo'}

        '''
        # pylint: disable-msg=E1101
        return {'slug': self.slug}

    @property
    def updated(self):
        '''Return the update time of the feed.

        If there are no posts, it is 1900-01-01 UTC, otherwise it will be the
        maximum update time across all posts.

        '''
        # pylint: disable-msg=E1101
        return max([post.updated for post in self.posts or []] +
                   [pytz.utc.localize(datetime.datetime(1900, 01, 01))])


class FeedParser(parsers.RstParser):
    r'''Parse all feeds matching ``[-a-zA-Z0-9_]+``.

    .. doctest::
       :hide:

       >>> environment['log'] = get_logger()

    .. doctest::

       >>> fp = FeedParser(environment, objects)
       >>> environment, objects, (parser,) = fp(environment, objects)
       >>> pprint(parser(environment, objects)) #doctest: +ELLIPSIS
       ({...},
        {'feed': [Feed(bar), Feed(baz), Feed(foo), Feed(quux)]},
        [])

    '''

    type = 'feed'
    paths = '^[-a-zA-Z0-9_.]+\.rst$'
    cls = Feed

    @decorators.in_environment('settings')
    def root(self, environment):
        '''The directory under which all feed objects reside.
        '''
        settings = environment['settings']
        return os.path.join(settings.CONTENT_ROOT, settings.FEEDS_SUBDIR)

    def rstparse(self, environment, objects, path, pieces):
        '''Use the parsed rst document to construct the necessary objects.
        '''
        attrs = {}
        attrs['slug'] = unicode(path[:-4])
        attrs['copyright'] = pieces['metadata'].get('copyright', '')
        attrs['_pub'] = pieces['pub']
        objects[self.type].append(self.cls(**attrs))


def _setup(test):
    '''Setup the tests.
    '''
    from pysettings.settings import Settings
    from firmant import routing
    test.globs['settings'] = Settings({'CONTENT_ROOT': 'testdata/pristine'
                                      ,'FEEDS_SUBDIR': 'feeds'
                                      })
    urlmapper = routing.URLMapper('/path/to/output/dir', 'http://testurl/', [])
    test.globs['environment'] = {'settings': test.globs['settings']
                                ,'urlmapper': urlmapper
                                }
    test.globs['objects'] = {}
